# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import math
from collections.abc import Sequence

from pyomo.common.autoslots import AutoSlots
from pyomo.common.numeric_types import check_if_numeric_type

try:
    from math import remainder
except ImportError:

    def remainder(a, b):
        ans = a % b
        if ans > abs(b / 2.0):
            ans -= b
        return ans


_inf = float('inf')
_infinite = {_inf, -_inf}


class RangeDifferenceError(ValueError):
    pass


class NumericRange(AutoSlots.Mixin):
    """A representation of a numeric range.

    This class represents a contiguous range of numbers.  The class
    mimics the Pyomo (*not* Python) `range` API, with a Start, End, and
    Step.  The Step is a signed int.  If the Step is 0, the range is
    continuous.  The End *is* included in the range.  Ranges are closed,
    unless `closed` is specified as a 2-tuple of bool values.  Only
    continuous ranges may be open (or partially open)

    Closed ranges are not necessarily strictly finite, as None is
    allowed for the End value (as well as the Start value, for
    continuous ranges only).

    Parameters
    ----------
        start : float
            The starting value for this NumericRange
        end : float
            The last value for this NumericRange
        step : int
            The interval between values in the range.  0 indicates a
            continuous range.  Negative values indicate discrete ranges
            walking backwards.
        closed : tuple of bool, optional
            A 2-tuple of bool values indicating if the beginning and end
            of the range is closed.  Open ranges are only allowed for
            continuous NumericRange objects.
    """

    __slots__ = ('start', 'end', 'step', 'closed')
    _EPS = 1e-15
    _types_comparable_to_int = {int}
    _closedMap = {
        True: True,
        False: False,
        '[': True,
        ']': True,
        '(': False,
        ')': False,
    }

    def __init__(self, start, end, step, closed=(True, True)):
        if int(step) != step:
            raise ValueError("NumericRange step must be int (got %s)" % (step,))
        step = int(step)
        if start is None:
            start = -_inf
        if end is None:
            end = math.copysign(_inf, step)

        if step:
            if start == -_inf:
                raise ValueError(
                    "NumericRange: start must not be None/-inf "
                    "for non-continuous steps"
                )
            if (end - start) * step < 0:
                raise ValueError(
                    "NumericRange: start, end ordering incompatible "
                    "with step direction (got [%s:%s:%s])" % (start, end, step)
                )
            if end not in _infinite:
                n = int((end - start) // step)
                new_end = start + n * step
                assert abs(end - new_end) < abs(step)
                end = new_end
                # It is important (for iterating) that all finite
                # discrete ranges have positive steps
                if step < 0:
                    start, end = end, start
                    step *= -1
        elif end < start:  # and step == 0
            raise ValueError(
                "NumericRange: start must be <= end for "
                "continuous ranges (got %s..%s)" % (start, end)
            )
        if start == end:
            # If this is a scalar, we will force the step to be 0 (so that
            # things like [1:5:10] == [1:50:100] are easier to validate)
            step = 0

        self.start = start
        self.end = end
        self.step = step

        self.closed = (self._closedMap[closed[0]], self._closedMap[closed[1]])
        if self.isdiscrete() and self.closed != (True, True):
            raise ValueError(
                "NumericRange %s is discrete, but passed closed=%s."
                "  Discrete ranges must be closed." % (self, self.closed)
            )

    def __str__(self):
        if not self.isdiscrete():
            return "%s%s..%s%s" % (
                "[" if self.closed[0] else "(",
                self.start,
                self.end,
                "]" if self.closed[1] else ")",
            )
        if self.start == self.end:
            return "[%s]" % (self.start,)
        elif self.step == 1:
            return "[%s:%s]" % (self.start, self.end)
        else:
            return "[%s:%s:%s]" % (self.start, self.end, self.step)

    __repr__ = __str__

    def __eq__(self, other):
        if type(other) is not NumericRange:
            return False
        return (
            self.start == other.start
            and self.end == other.end
            and self.step == other.step
            and self.closed == other.closed
        )

    def __ne__(self, other):
        return not self.__eq__(other)

    def __contains__(self, value):
        # NumericRanges must hold items that are comparable to ints
        if value.__class__ not in self._types_comparable_to_int:
            # Build on numeric_type.check_if_numeric_type to cleanly
            # handle numpy registrations
            if check_if_numeric_type(value):
                self._types_comparable_to_int.add(value.__class__)
            else:
                # Special case: because numpy is fond of returning scalars
                # as length-1 ndarrays, we will include a special case that
                # will unpack things that look like single element arrays.
                try:
                    # Note: trap "value[0] is not value" to catch things like
                    # single-character strings
                    if (
                        hasattr(value, '__len__')
                        and hasattr(value, '__getitem__')
                        and len(value) == 1
                        and value[0] is not value
                    ):
                        return value[0] in self
                except:
                    pass
                # See if this class behaves like a "normal" number: both
                # comparable and creatable
                try:
                    if not (bool(value - 0 > 0) ^ bool(value - 0 <= 0)):
                        return False
                    elif value.__class__(0) != 0 or not value.__class__(0) == 0:
                        return False
                    else:
                        self._types_comparable_to_int.add(value.__class__)
                except:
                    return False

        if self.step:
            _dir = int(math.copysign(1, self.step))
            _from_start = value - self.start
            return (
                0 <= _dir * _from_start <= _dir * (self.end - self.start)
                and abs(remainder(_from_start, self.step)) <= self._EPS
            )
        else:
            return (value >= self.start if self.closed[0] else value > self.start) and (
                value <= self.end if self.closed[1] else value < self.end
            )

    @staticmethod
    def _continuous_discrete_disjoint(cont, disc):
        # At this point, we know the ranges overlap (tested for at the
        # beginning of isdisjoint()
        d_lb = disc.start if disc.step > 0 else disc.end
        d_ub = disc.end if disc.step > 0 else disc.start
        if cont.start <= d_lb:
            return False
        if cont.end >= d_ub:
            return False

        EPS = NumericRange._EPS
        if cont.end - cont.start - EPS > abs(disc.step):
            return False
        # At this point, the continuous set is shorter than the discrete
        # step.  We need to see if the continuous set overlaps one of the
        # points, or lies completely between two.
        #
        # Note, taking the absolute value of the step is safe because we are
        # seeing if the continuous range overlaps a discrete point in the
        # underlying (unbounded) discrete sequence grounded by disc.start
        rStart = remainder(cont.start - disc.start, abs(disc.step))
        rEnd = remainder(cont.end - disc.start, abs(disc.step))
        return (
            (abs(rStart) > EPS or not cont.closed[0])
            and (abs(rEnd) > EPS or not cont.closed[1])
            and (rStart - rEnd > 0 or not any(cont.closed))
        )

    def isdiscrete(self):
        return self.step or self.start == self.end

    def isfinite(self):
        return (self.step and self.end not in _infinite) or self.end == self.start

    def isdisjoint(self, other):
        if not isinstance(other, NumericRange):
            # It is easier to just use NonNumericRange/AnyRange's
            # implementation
            return other.isdisjoint(self)

        # First, do a simple sanity check on the endpoints
        if self._nooverlap(other):
            return True
        # Now, check continuous sets
        if not self.step or not other.step:
            # Check the case of scalar values
            if self.start == self.end:
                return self.start not in other
            elif other.start == other.end:
                return other.start not in self

            # We now need to check a continuous set is a subset of a discrete
            # set and the continuous set sits between discrete points
            if self.step:
                return NumericRange._continuous_discrete_disjoint(other, self)
            elif other.step:
                return NumericRange._continuous_discrete_disjoint(self, other)
            else:
                # 2 continuous sets, with overlapping end points: not disjoint
                return False
        # both sets are discrete
        if self.step == other.step:
            return abs(remainder(other.start - self.start, self.step)) > self._EPS
        # Two infinite discrete sets will *eventually* have a common
        # point.  This is trivial for coprime integer steps.  For steps
        # with gcd > 1, we need to ensure that the two ranges are
        # aligned to the gcd period.
        #
        # Note that this all breaks down for for float steps with
        # infinite precision (think a step of PI).  However, for finite
        # precision maths, the "float" times a sufficient power of two
        # is an integer.  Is this a distinction we want to make?
        # Personally, anyone making a discrete set with a non-integer
        # step is asking for trouble.  Maybe the better solution is to
        # require that the step be integer (which is what we do).
        elif (
            self.end in _infinite
            and other.end in _infinite
            and self.step * other.step > 0
        ):
            gcd = NumericRange._gcd(self.step, other.step)
            return abs(remainder(other.start - self.start, gcd)) > self._EPS
        # OK - at this point, there are a finite number of set members
        # that can overlap.  Just check all the members of one set
        # against the other
        if self.step > 0:
            end = min(self.end, max(other.start, other.end))
        else:
            end = max(self.end, min(other.start, other.end))
        i = 0
        item = self.start
        while (self.step > 0 and item <= end) or (self.step < 0 and item >= end):
            if item in other:
                return False
            i += 1
            item = self.start + self.step * i
        return True

    def issubset(self, other):
        if not isinstance(other, NumericRange):
            if type(other) is AnyRange:
                return True
            elif type(other) is NonNumericRange:
                return False
            # Other non NumericRange objects will generate
            # AttributeError exceptions below

        # First, do a simple sanity check on the endpoints
        s1, e1, c1 = self.normalize_bounds()
        s2, e2, c2 = other.normalize_bounds()
        # Checks for unbounded ranges and to make sure self's endpoints are
        # within other's endpoints.
        if s1 < s2:
            return False
        if e1 > e2:
            return False
        if s1 == s2 and c1[0] and not c2[0]:
            return False
        if e1 == e2 and c1[1] and not c2[1]:
            return False
        # If other is continuous (even a single point), then by
        # definition, self is a subset (regardless of step)
        if other.step == 0:
            return True
        # If other is discrete and self is continuous, then self can't be a
        # subset unless self is a scalar and is in other
        elif self.step == 0:
            if self.start == self.end:
                return self.start in other
            else:
                return False
        # At this point, both sets are discrete.  Self's period must be a
        # positive integer multiple of other's ...
        EPS = NumericRange._EPS
        if abs(remainder(self.step, other.step)) > EPS:
            return False
        # ...and they must shart a point in common
        return abs(remainder(other.start - self.start, other.step)) <= EPS

    def normalize_bounds(self):
        """Normalizes this NumericRange.

        This returns a normalized range by reversing lb and ub if the
        NumericRange step is less than zero.  If lb and ub are
        reversed, then closed is updated to reflect that change.

        Returns
        -------
        lb, ub, closed

        """
        if self.step >= 0:
            return self.start, self.end, self.closed
        else:
            return self.end, self.start, (self.closed[1], self.closed[0])

    def _nooverlap(self, other):
        """Return True if the ranges for self and other are strictly separate"""
        s1, e1, c1 = self.normalize_bounds()
        s2, e2, c2 = other.normalize_bounds()
        if (
            e1 < s2
            or e2 < s1
            or (e1 == s2 and not (c1[1] and c2[0]))
            or (e2 == s1 and not (c2[1] and c1[0]))
        ):
            return True
        return False

    @staticmethod
    def _split_ranges(cnr, new_step):
        """Split a discrete range into a list of ranges using a new step.

        This takes a single NumericRange and splits it into a set
        of new ranges, all of which use a new step.  The new_step must
        be a multiple of the current step.  CNR objects with a step of 0
        are returned unchanged.

        Parameters
        ----------
            cnr: `NumericRange`
                The range to split
            new_step: `int`
                The new step to use for returned ranges

        """
        if cnr.step == 0 or new_step == 0:
            return [cnr]

        assert new_step >= abs(cnr.step)
        assert new_step % cnr.step == 0
        _dir = int(math.copysign(1, cnr.step))
        _subranges = []
        for i in range(int(abs(new_step // cnr.step))):
            if _dir * (cnr.start + i * cnr.step) > _dir * cnr.end:
                # Once we walk past the end of the range, we are done
                # (all remaining offsets will be farther past the end)
                break
            _subranges.append(
                NumericRange(cnr.start + i * cnr.step, cnr.end, _dir * new_step)
            )
        return _subranges

    @staticmethod
    def _gcd(a, b):
        while b != 0:
            a, b = b, a % b
        return a

    @staticmethod
    def _lcm(a, b):
        gcd = NumericRange._gcd(a, b)
        if not gcd:
            return 0
        return a * b / gcd

    def _step_lcm(self, other_ranges):
        """This computes an approximate Least Common Multiple step"""
        # Note: scalars are discrete, but have a step of 0.  Pretend the
        # step is 1 so that we can compute a realistic "step lcm"
        if self.isdiscrete():
            a = self.step or 1
        else:
            a = 0
        for o in other_ranges:
            if o.isdiscrete():
                b = o.step or 1
            else:
                b = 0
            lcm = NumericRange._lcm(a, b)
            # This is a modified LCM.  LCM(n,0) == 0, but for step
            # calculations, we want it to be n
            if lcm:
                a = lcm
            else:
                # one of the steps was 0: add to preserve the non-zero step
                a += b
        return int(abs(a))

    def _push_to_discrete_element(self, val, push_to_next_larger_value):
        if not self.step or val in _infinite:
            return val
        else:
            # self is discrete and val is a numeric value.  Move val to
            # the first discrete point aligned with self's range
            #
            # Note that we need to push the value INTO the range defined
            # by this set, so floor/ceil depends on the sign of self.step
            if push_to_next_larger_value:
                _rndFcn = math.ceil if self.step > 0 else math.floor
            else:
                _rndFcn = math.floor if self.step > 0 else math.ceil
            return self.start + self.step * _rndFcn(
                (val - self.start) / float(self.step)
            )

    def range_difference(self, other_ranges):
        """Return the difference between this range and a list of other ranges.

        Parameters
        ----------
            other_ranges: `iterable`
                An iterable of other range objects to subtract from this range

        """
        _cnr_other_ranges = []
        for r in other_ranges:
            if isinstance(r, NumericRange):
                _cnr_other_ranges.append(r)
            elif type(r) is AnyRange:
                return []
            elif type(r) is NonNumericRange:
                continue
            else:
                # Note: important to check and raise an exception here;
                # otherwise, unrecognized range types would be silently
                # ignored.
                raise ValueError("Unknown range type, %s" % (type(r).__name__,))

        other_ranges = _cnr_other_ranges

        # Find the Least Common Multiple of all the range steps.  We
        # will split discrete ranges into separate ranges with this step
        # so that we can more easily compare them.
        lcm = self._step_lcm(other_ranges)

        # Split this range into subranges
        _this = NumericRange._split_ranges(self, lcm)
        # Split the other range(s) into subranges
        _other = []
        for s in other_ranges:
            _other.extend(NumericRange._split_ranges(s, lcm))
        # For each rhs subrange, s
        for s in _other:
            _new_subranges = []
            for t in _this:
                if t._nooverlap(s):
                    # If s and t have no overlap, then s cannot remove
                    # any elements from t
                    _new_subranges.append(t)
                    continue

                if t.isdiscrete():
                    # s and t are discrete ranges.  Note if there is a
                    # discrete range in the list of ranges, then lcm > 0
                    if s.isdiscrete() and (s.start - t.start) % lcm != 0:
                        # s is offset from t and cannot remove any
                        # elements
                        _new_subranges.append(t)
                        continue

                t_min, t_max, t_c = t.normalize_bounds()
                s_min, s_max, s_c = s.normalize_bounds()

                if s.isdiscrete() and not t.isdiscrete():
                    #
                    # This handles the special case of continuous-discrete
                    if (s_min == -_inf and t.start == -_inf) or (
                        s_max == _inf and t.end == _inf
                    ):
                        raise RangeDifferenceError(
                            "We do not support subtracting an infinite "
                            "discrete range %s from an infinite continuous "
                            "range %s" % (s, t)
                        )

                    # At least one of s_min amd t.start must be non-inf
                    start = max(s_min, s._push_to_discrete_element(t.start, True))
                    # At least one of s_max amd t.end must be non-inf
                    end = min(s_max, s._push_to_discrete_element(t.end, False))

                    if t.start < start:
                        _new_subranges.append(
                            NumericRange(t.start, start, 0, (t.closed[0], False))
                        )
                    if s.step:  # i.e., not a single point
                        for i in range(int((end - start) // s.step)):
                            _new_subranges.append(
                                NumericRange(
                                    start + i * s.step,
                                    start + (i + 1) * s.step,
                                    0,
                                    '()',
                                )
                            )
                    if t.end > end:
                        _new_subranges.append(
                            NumericRange(end, t.end, 0, (False, t.closed[1]))
                        )
                else:
                    #
                    # This handles discrete-discrete,
                    # continuous-continuous, and discrete-continuous
                    #
                    if t_min < s_min:
                        # Note s_min will never be -inf due to the < test
                        if t.step:
                            s_min -= lcm
                            closed1 = True
                        _min = min(t_max, s_min)
                        if not t.step:
                            closed1 = not s_c[0] if _min is s_min else t_c[1]
                        _closed = (t_c[0], closed1)
                        _step = abs(t.step)
                        _rng = t_min, _min
                        if t_min == -_inf and t.step:
                            _step = -_step
                            _rng = _rng[1], _rng[0]
                            _closed = _closed[1], _closed[0]

                        _new_subranges.append(
                            NumericRange(_rng[0], _rng[1], _step, _closed)
                        )
                    elif t_min == s_min and t_c[0] and not s_c[0]:
                        _new_subranges.append(NumericRange(t_min, t_min, 0))

                    if t_max > s_max:
                        # Note s_max will never be inf due to the > test
                        if t.step:
                            s_max += lcm
                            closed0 = True
                        _max = max(t_min, s_max)
                        if not t.step:
                            closed0 = not s_c[1] if _max is s_max else t_c[0]
                        _new_subranges.append(
                            NumericRange(_max, t_max, abs(t.step), (closed0, t_c[1]))
                        )
                    elif t_max == s_max and t_c[1] and not s_c[1]:
                        _new_subranges.append(NumericRange(t_max, t_max, 0))
            _this = _new_subranges
        return _this

    def range_intersection(self, other_ranges):
        """Return the intersection between this range and a set of other ranges.

        Parameters
        ----------
            other_ranges: `iterable`
                An iterable of other range objects to intersect with this range

        """
        _cnr_other_ranges = []
        for r in other_ranges:
            if isinstance(r, NumericRange):
                _cnr_other_ranges.append(r)
            elif type(r) is AnyRange:
                return [self]
            elif type(r) is NonNumericRange:
                continue
            else:
                # Note: important to check and raise an exception here;
                # otherwise, unrecognized range types would be silently
                # ignored.
                raise ValueError("Unknown range type, %s" % (type(r).__name__,))
        other_ranges = _cnr_other_ranges

        # Find the Least Common Multiple of all the range steps.  We
        # will split discrete ranges into separate ranges with this step
        # so that we can more easily compare them.
        lcm = self._step_lcm(other_ranges)

        ans = []
        # Split this range into subranges
        _this = NumericRange._split_ranges(self, lcm)
        # Split the other range(s) into subranges
        _other = []
        for s in other_ranges:
            _other.extend(NumericRange._split_ranges(s, lcm))
        # For each lhs subrange, t
        for t in _this:
            # Compare it against each rhs range and only keep the
            # subranges of this range that are inside the lhs range
            for s in _other:
                if s.isdiscrete() and t.isdiscrete():
                    # s and t are discrete ranges.  Note if there is a
                    # finite range in the list of ranges, then lcm > 0
                    if (s.start - t.start) % lcm != 0:
                        # s is offset from t and cannot have any
                        # elements in common
                        continue
                if t._nooverlap(s):
                    continue

                t_min, t_max, t_c = t.normalize_bounds()
                s_min, s_max, s_c = s.normalize_bounds()
                step = abs(t.step if t.step else s.step)

                intersect_start = max(
                    t._push_to_discrete_element(s_min, True),
                    s._push_to_discrete_element(t_min, True),
                )

                intersect_end = min(
                    t._push_to_discrete_element(s_max, False),
                    s._push_to_discrete_element(t_max, False),
                )
                c = [True, True]
                if intersect_start == t_min:
                    c[0] &= t_c[0]
                if intersect_start == s_min:
                    c[0] &= s_c[0]
                if intersect_end == t_max:
                    c[1] &= t_c[1]
                if intersect_end == s_max:
                    c[1] &= s_c[1]
                if step and intersect_start == -_inf:
                    ans.append(
                        NumericRange(
                            intersect_end, intersect_start, -step, (c[1], c[0])
                        )
                    )
                else:
                    ans.append(NumericRange(intersect_start, intersect_end, step, c))
        return ans


class NonNumericRange:
    """A range-like object for representing a single non-numeric value

    The class name is a bit of a misnomer, as this object does not
    represent a range but rather a single value.  However, as it
    duplicates the Range API (as used by :py:class:`NumericRange`), it
    is called a "Range".

    """

    __slots__ = ('value',)

    def __init__(self, val):
        self.value = val

    def __str__(self):
        return "{%s}" % (self.value,)

    __repr__ = __str__

    def __eq__(self, other):
        return isinstance(other, NonNumericRange) and other.value == self.value

    def __ne__(self, other):
        return not self.__eq__(other)

    def __contains__(self, value):
        return value == self.value

    def __getstate__(self):
        """
        Retrieve the state of this object as a dictionary.

        This method must be defined because this class uses slots.
        """
        state = {}  # super(NonNumericRange, self).__getstate__()
        for i in NonNumericRange.__slots__:
            state[i] = getattr(self, i)
        return state

    def __setstate__(self, state):
        """
        Set the state of this object using values from a state dictionary.

        This method must be defined because this class uses slots.
        """
        for key, val in state.items():
            # Note: per the Python data model docs, we explicitly
            # set the attribute using object.__setattr__() instead
            # of setting self.__dict__[key] = val.
            object.__setattr__(self, key, val)

    def isdiscrete(self):
        return True

    def isfinite(self):
        return True

    def isdisjoint(self, other):
        return self.value not in other

    def issubset(self, other):
        return self.value in other

    def range_difference(self, other_ranges):
        for r in other_ranges:
            if self.value in r:
                return []
        return [self]

    def range_intersection(self, other_ranges):
        for r in other_ranges:
            if self.value in r:
                return [self]
        return []


class AnyRange:
    """A range object for representing Any sets"""

    __slots__ = ()

    def __init__(self):
        pass

    def __str__(self):
        return "[*]"

    __repr__ = __str__

    def __eq__(self, other):
        return isinstance(other, AnyRange)

    def __ne__(self, other):
        return not self.__eq__(other)

    def __contains__(self, value):
        return True

    def isdiscrete(self):
        return False

    def isfinite(self):
        return False

    def isdisjoint(self, other):
        return False

    def issubset(self, other):
        return isinstance(other, AnyRange)

    def range_difference(self, other_ranges):
        for r in other_ranges:
            if isinstance(r, AnyRange):
                return []
        else:
            return [self]

    def range_intersection(self, other_ranges):
        return list(other_ranges)


class RangeProduct:
    """A range-like object for representing the cross product of ranges"""

    __slots__ = ('range_lists',)

    def __init__(self, range_lists):
        self.range_lists = range_lists
        # Type checking.  Users should never create a RangeProduct, but
        # just in case...
        assert range_lists.__class__ is list
        for subrange in range_lists:
            assert subrange.__class__ is list

    def __str__(self):
        return (
            "<"
            + ', '.join(
                str(tuple(_)) if len(_) > 1 else str(_[0]) for _ in self.range_lists
            )
            + ">"
        )

    __repr__ = __str__

    def __eq__(self, other):
        return (
            isinstance(other, RangeProduct)
            and self.range_difference([other]) == []
            and other.range_difference([self]) == []
        )

    def __ne__(self, other):
        return not self.__eq__(other)

    def __contains__(self, value):
        if not isinstance(value, Sequence):
            return False
        if len(value) != len(self.range_lists):
            return False
        return all(
            any(val in rng for rng in rng_list)
            for val, rng_list in zip(value, self.range_lists)
        )

    def __getstate__(self):
        """
        Retrieve the state of this object as a dictionary.

        This method must be defined because this class uses slots.
        """
        state = {}  # super(RangeProduct, self).__getstate__()
        for i in RangeProduct.__slots__:
            state[i] = getattr(self, i)
        return state

    def __setstate__(self, state):
        """
        Set the state of this object using values from a state dictionary.

        This method must be defined because this class uses slots.
        """
        for key, val in state.items():
            # Note: per the Python data model docs, we explicitly
            # set the attribute using object.__setattr__() instead
            # of setting self.__dict__[key] = val.
            object.__setattr__(self, key, val)

    def isdiscrete(self):
        return all(
            all(rng.isdiscrete() for rng in rng_list) for rng_list in self.range_lists
        )

    def isfinite(self):
        return all(
            all(rng.isfinite() for rng in rng_list) for rng_list in self.range_lists
        )

    def isdisjoint(self, other):
        if type(other) is AnyRange:
            return False
        if type(other) is not RangeProduct:
            return True
        if len(other.range_lists) != len(self.range_lists):
            return True
        # Remember, range_lists is a list of lists of range objects.  As
        # isdisjoint only accepts range objects, we need to unpack
        # everything.  Non-disjoint range products require overlaps in
        # all dimensions.
        for s, o in zip(self.range_lists, other.range_lists):
            if all(s_rng.isdisjoint(o_rng) for s_rng in s for o_rng in o):
                return True
        return False

    def issubset(self, other):
        if type(other) is AnyRange:
            return True
        return not any(_ for _ in self.range_difference([other]))

    def range_difference(self, other_ranges):
        # The goal is to start with a single range product and create a
        # set of range products that, when combined, model the
        # range_difference.  This will potentially create (redundant)
        # overlapping regions, but that is OK.
        ans = [self]
        N = len(self.range_lists)
        for other in other_ranges:
            if type(other) is AnyRange:
                return []
            if type(other) is not RangeProduct or len(other.range_lists) != N:
                continue

            tmp = []
            for rp in ans:
                if rp.isdisjoint(other):
                    tmp.append(rp)
                    continue

                for dim in range(N):
                    remainder = []
                    for r in rp.range_lists[dim]:
                        remainder.extend(r.range_difference(other.range_lists[dim]))
                    if remainder:
                        tmp.append(RangeProduct(list(rp.range_lists)))
                        tmp[-1].range_lists[dim] = remainder
            ans = tmp
        return ans

    def range_intersection(self, other_ranges):
        # The goal is to start with a single range product and create a
        # set of range products that, when combined, model the
        # range_difference.  This will potentially create (redundant)
        # overlapping regions, but that is OK.
        ans = list(self.range_lists)
        N = len(self.range_lists)
        for other in other_ranges:
            if type(other) is AnyRange:
                continue
            if type(other) is not RangeProduct or len(other.range_lists) != N:
                return []

            for dim in range(N):
                tmp = []
                for r in ans[dim]:
                    tmp.extend(r.range_intersection(other.range_lists[dim]))
                if not tmp:
                    return []
                ans[dim] = tmp
        return [RangeProduct(ans)]



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import pyomo.environ as pyo
import pyomo.dae as dae
from pyomo.common.dependencies import networkx_available
from pyomo.common.dependencies import scipy_available
from pyomo.common.collections import ComponentSet, ComponentMap
from pyomo.contrib.incidence_analysis.interface import get_structural_incidence_matrix
from pyomo.contrib.incidence_analysis.dulmage_mendelsohn import dulmage_mendelsohn
from pyomo.contrib.incidence_analysis.tests.models_for_testing import (
    make_gas_expansion_model,
    make_dynamic_model,
)

import pyomo.common.unittest as unittest


@unittest.skipUnless(networkx_available, "networkx is not available.")
@unittest.skipUnless(scipy_available, "scipy is not available.")
class TestGasExpansionDMMatrixInterface(unittest.TestCase):
    def test_square_well_posed_model(self):
        N = 4
        m = make_gas_expansion_model(N)
        m.F[0].fix()
        m.rho[0].fix()
        m.T[0].fix()

        variables = [v for v in m.component_data_objects(pyo.Var) if not v.fixed]
        constraints = list(m.component_data_objects(pyo.Constraint))
        imat = get_structural_incidence_matrix(variables, constraints)

        N, M = imat.shape
        self.assertEqual(N, M)

        row_partition, col_partition = dulmage_mendelsohn(imat)

        # Unmatched, reachable-from-unmatched, and matched-with-reachable
        # subsets are all empty.
        self.assertEqual(len(row_partition[0]), 0)
        self.assertEqual(len(row_partition[1]), 0)
        self.assertEqual(len(row_partition[2]), 0)
        self.assertEqual(len(col_partition[0]), 0)
        self.assertEqual(len(col_partition[1]), 0)
        self.assertEqual(len(col_partition[2]), 0)

        # All nodes belong to the "well-determined" subset
        self.assertEqual(len(row_partition[3]), M)
        self.assertEqual(len(col_partition[3]), N)

    def test_square_ill_posed_model(self):
        N = 1
        m = make_gas_expansion_model(N)
        m.P[0].fix()
        m.rho[0].fix()
        m.T[0].fix()

        variables = [v for v in m.component_data_objects(pyo.Var) if not v.fixed]
        constraints = list(m.component_data_objects(pyo.Constraint))
        imat = get_structural_incidence_matrix(variables, constraints)

        var_idx_map = ComponentMap((v, i) for i, v in enumerate(variables))
        con_idx_map = ComponentMap((c, i) for i, c in enumerate(constraints))

        N, M = imat.shape
        self.assertEqual(N, M)

        row_partition, col_partition = dulmage_mendelsohn(imat)

        # Only unmatched constraint is ideal_gas[0]
        unmatched_rows = [con_idx_map[m.ideal_gas[0]]]
        self.assertEqual(row_partition[0], unmatched_rows)
        # No other constraints can possibly be unmatched.
        self.assertEqual(row_partition[1], [])
        # The potentially unmatched variables have four constraints
        # between them
        matched_con_set = set(
            con_idx_map[con] for con in constraints if con is not m.ideal_gas[0]
        )
        self.assertEqual(set(row_partition[2]), matched_con_set)

        # All variables are potentially unmatched
        potentially_unmatched_set = set(range(len(variables)))
        potentially_unmatched = col_partition[0] + col_partition[1]
        self.assertEqual(set(potentially_unmatched), potentially_unmatched_set)

    def test_rectangular_system(self):
        N_model = 2
        m = make_gas_expansion_model(N_model)
        variables = list(m.component_data_objects(pyo.Var))
        constraints = list(m.component_data_objects(pyo.Constraint))

        imat = get_structural_incidence_matrix(variables, constraints)
        M, N = imat.shape
        self.assertEqual(M, 4 * N_model + 1)
        self.assertEqual(N, 4 * (N_model + 1))

        row_partition, col_partition = dulmage_mendelsohn(imat)

        # No unmatched constraints
        self.assertEqual(row_partition[0], [])
        self.assertEqual(row_partition[1], [])

        # All constraints are matched with potentially unmatched variables
        matched_con_set = set(range(len(constraints)))
        self.assertEqual(set(row_partition[2]), matched_con_set)

        # 3 unmatched variables
        self.assertEqual(len(col_partition[0]), 3)
        # All variables are potentially unmatched
        potentially_unmatched = col_partition[0] + col_partition[1]
        potentially_unmatched_set = set(range(len(variables)))
        self.assertEqual(set(potentially_unmatched), potentially_unmatched_set)

    def test_recover_matching(self):
        N_model = 4
        m = make_gas_expansion_model(N_model)
        variables = list(m.component_data_objects(pyo.Var))
        constraints = list(m.component_data_objects(pyo.Constraint))
        imat = get_structural_incidence_matrix(variables, constraints)
        rdmp, cdmp = dulmage_mendelsohn(imat)
        rmatch = rdmp.underconstrained + rdmp.square + rdmp.overconstrained
        cmatch = cdmp.underconstrained + cdmp.square + cdmp.overconstrained
        matching = list(zip(rmatch, cmatch))
        rmatch = [r for (r, c) in matching]
        cmatch = [c for (r, c) in matching]
        # Assert that the matched rows and columns contain no duplicates
        self.assertEqual(len(set(rmatch)), len(rmatch))
        self.assertEqual(len(set(cmatch)), len(cmatch))
        entry_set = set(zip(imat.row, imat.col))
        for i, j in matching:
            # Assert that each pair in the matching is a valid entry
            # in the matrix
            self.assertIn((i, j), entry_set)


@unittest.skipUnless(networkx_available, "networkx is not available.")
@unittest.skipUnless(scipy_available, "scipy is not available.")
class TestDynamicModel(unittest.TestCase):
    def test_rectangular_model(self):
        m = make_dynamic_model()

        m.height[0].fix()

        variables = [v for v in m.component_data_objects(pyo.Var) if not v.fixed]
        constraints = list(m.component_data_objects(pyo.Constraint))

        imat = get_structural_incidence_matrix(variables, constraints)
        M, N = imat.shape
        var_idx_map = ComponentMap((v, i) for i, v in enumerate(variables))
        con_idx_map = ComponentMap((c, i) for i, c in enumerate(constraints))

        row_partition, col_partition = dulmage_mendelsohn(imat)

        # No unmatched rows
        self.assertEqual(row_partition[0], [])
        self.assertEqual(row_partition[1], [])

        # Assert that the square subsystem contains the components we expect
        self.assertEqual(len(row_partition[3]), 1)
        self.assertEqual(row_partition[3][0], con_idx_map[m.flow_out_eqn[0]])

        self.assertEqual(len(col_partition[3]), 1)
        self.assertEqual(col_partition[3][0], var_idx_map[m.flow_out[0]])

        # Assert that underdetermined subsystem contains the components
        # we expect

        # Rows matched with potentially unmatched columns
        self.assertEqual(len(row_partition[2]), M - 1)
        row_indices = set([i for i in range(M) if i != con_idx_map[m.flow_out_eqn[0]]])
        self.assertEqual(set(row_partition[2]), row_indices)

        # Potentially unmatched columns
        self.assertEqual(len(col_partition[0]), N - M)
        self.assertEqual(len(col_partition[1]), M - 1)
        potentially_unmatched = col_partition[0] + col_partition[1]
        col_indices = set([i for i in range(N) if i != var_idx_map[m.flow_out[0]]])
        self.assertEqual(set(potentially_unmatched), col_indices)


if __name__ == "__main__":
    unittest.main()



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import logging
import types

from pyomo.common.log import is_debug_set
from pyomo.common.timing import ConstructionTimer
from pyomo.core.base.component import ModelComponentFactory
from pyomo.core.base.indexed_component import IndexedComponent
from pyomo.core.base.misc import apply_indexed_rule

logger = logging.getLogger('pyomo.core')


@ModelComponentFactory.register(
    "A component that performs tests during model construction.  The action rule is applied to every index value."
)
class BuildCheck(IndexedComponent):
    """
    A build check, which executes a rule for all valid indices.  If
    the function returns False an exception is raised.

    Constructor arguments:
        rule         The rule that is executed for every indice.

    Private class attributes:
        _rule       The rule that is executed for every indice.
    """

    def __init__(self, *args, **kwd):
        self._rule = kwd.pop('rule', None)
        kwd['ctype'] = BuildCheck
        IndexedComponent.__init__(self, *args, **kwd)
        #
        if not type(self._rule) is types.FunctionType:
            raise ValueError(
                "BuildCheck  must have an 'rule' option specified whose value is a function"
            )

    def _pprint(self):
        return ([], None, None, None)

    def construct(self, data=None):
        """Apply the rule to construct values in this set"""
        if is_debug_set(logger):  # pragma:nocover
            logger.debug("Constructing Check, name=" + self.name)
        #
        if self._constructed:  # pragma:nocover
            return
        timer = ConstructionTimer(self)
        self._constructed = True
        #
        if not self.is_indexed():
            # Scalar component
            res = self._rule(self._parent())
            if not res:
                raise ValueError("BuildCheck %r identified error" % self.name)
        else:
            # Indexed component
            for index in self._index_set:
                res = apply_indexed_rule(self, self._rule, self._parent(), index)
                if not res:
                    raise ValueError(
                        "BuildCheck %r identified error with index %r"
                        % (self.name, str(index))
                    )
        timer.report()



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

"""
Methods for execution of the main PyROS cutting set algorithm.
"""

from collections import namedtuple

from pyomo.common.dependencies import numpy as np
from pyomo.common.collections import ComponentMap
from pyomo.core.base import value

import pyomo.contrib.pyros.master_problem_methods as mp_methods
import pyomo.contrib.pyros.separation_problem_methods as sp_methods
from pyomo.contrib.pyros.util import (
    check_time_limit_reached,
    ObjectiveType,
    pyrosTerminationCondition,
    IterationLogRecord,
    get_main_elapsed_time,
    get_dr_var_to_monomial_map,
)


class GRCSResults:
    """
    Cutting set RO algorithm solve results.

    Attributes
    ----------
    master_results : MasterResults
        Solve results for most recent master problem.
    separation_results : SeparationResults or None
        Solve results for separation problem(s) of last iteration.
        If the separation subroutine was not invoked in the last
        iteration, then None.
    pyros_termination_condition : pyrosTerminationCondition
        PyROS termination condition.
    iterations : int
        Number of iterations required.
    """

    def __init__(
        self,
        master_results,
        separation_results,
        pyros_termination_condition,
        iterations,
    ):
        self.master_results = master_results
        self.separation_results = separation_results
        self.pyros_termination_condition = pyros_termination_condition
        self.iterations = iterations


def _evaluate_shift(current, prev, initial, norm=None):
    if current.size == 0:
        return None
    else:
        normalizers = np.max(
            np.vstack((np.ones(initial.size), np.abs(initial))), axis=0
        )
        return np.max(np.abs(current - prev) / normalizers)


VariableValueData = namedtuple(
    "VariableValueData",
    ("first_stage_variables", "second_stage_variables", "decision_rule_monomials"),
)


def get_variable_value_data(working_blk, dr_var_to_monomial_map):
    """
    Get variable value data.
    """
    ep = working_blk.effective_var_partitioning

    first_stage_data = ComponentMap(
        (var, var.value) for var in ep.first_stage_variables
    )
    second_stage_data = ComponentMap(
        (var, var.value) for var in ep.second_stage_variables
    )
    dr_term_data = ComponentMap(
        (dr_var, value(monomial))
        for dr_var, monomial in get_dr_var_to_monomial_map(working_blk).items()
    )

    return VariableValueData(
        first_stage_variables=first_stage_data,
        second_stage_variables=second_stage_data,
        decision_rule_monomials=dr_term_data,
    )


def evaluate_variable_shifts(current_var_data, previous_var_data, initial_var_data):
    """
    Evaluate relative changes in the variable values
    across solutions to a working model block, such as the
    nominal master block.
    """
    if previous_var_data is None:
        return None, None, None
    else:
        var_shifts = []
        for attr in current_var_data._fields:
            var_shifts.append(
                _evaluate_shift(
                    current=np.array(list(getattr(current_var_data, attr).values())),
                    prev=np.array(list(getattr(previous_var_data, attr).values())),
                    initial=np.array(list(getattr(initial_var_data, attr).values())),
                )
            )

    return tuple(var_shifts)


def ROSolver_iterative_solve(model_data):
    """
    Solve an RO problem with the iterative GRCS algorithm.

    Parameters
    ----------
    model_data : model data object
        Model data object, equipped with the
        fully preprocessed working model.

    Returns
    -------
    GRCSResults
        Iterative solve results.
    """
    config = model_data.config
    master_data = mp_methods.MasterProblemData(model_data)
    separation_data = sp_methods.SeparationProblemData(model_data)

    # set up first-stage variable and DR variable sets
    nominal_master_blk = master_data.master_model.scenarios[0, 0]
    dr_var_monomial_map = get_dr_var_to_monomial_map(nominal_master_blk)

    # keep track of variable values for iteration logging
    first_iter_var_data = None
    previous_iter_var_data = None
    current_iter_var_data = None

    num_second_stage_ineq_cons = len(
        separation_data.separation_model.second_stage.inequality_cons
    )
    IterationLogRecord.log_header(config.progress_logger.info)
    k = 0
    while config.max_iter == -1 or k < config.max_iter:
        master_data.iteration = k
        config.progress_logger.debug(f"PyROS working on iteration {k}...")

        master_soln = master_data.solve_master()
        master_termination_not_acceptable = master_soln.pyros_termination_condition in {
            pyrosTerminationCondition.robust_infeasible,
            pyrosTerminationCondition.time_out,
            pyrosTerminationCondition.subsolver_error,
        }
        if master_termination_not_acceptable:
            iter_log_record = IterationLogRecord(
                iteration=k,
                objective=None,
                first_stage_var_shift=None,
                second_stage_var_shift=None,
                dr_var_shift=None,
                num_violated_cons=None,
                max_violation=None,
                dr_polishing_success=None,
                all_sep_problems_solved=None,
                global_separation=None,
                elapsed_time=get_main_elapsed_time(model_data.timing),
                master_backup_solver=master_soln.backup_solver_used,
                master_feasibility_success=master_soln.feasibility_problem_success,
                separation_backup_local_solver=None,
                separation_backup_global_solver=None,
            )
            iter_log_record.log(config.progress_logger.info)
            return GRCSResults(
                master_results=master_soln,
                separation_results=None,
                pyros_termination_condition=master_soln.pyros_termination_condition,
                iterations=k + 1,
            )

        polishing_successful = True
        polish_master_solution = (
            config.decision_rule_order != 0
            and nominal_master_blk.first_stage.decision_rule_vars
            and k != 0
        )
        if polish_master_solution:
            _, polishing_successful = master_data.solve_dr_polishing()

        # track variable values
        current_iter_var_data = get_variable_value_data(
            nominal_master_blk, dr_var_monomial_map
        )
        if k == 0:
            first_iter_var_data = current_iter_var_data
            previous_iter_var_data = None

        fsv_shift, ssv_shift, dr_var_shift = evaluate_variable_shifts(
            current_var_data=current_iter_var_data,
            previous_var_data=previous_iter_var_data,
            initial_var_data=first_iter_var_data,
        )

        # === Check if time limit reached after polishing
        if check_time_limit_reached(model_data.timing, config):
            iter_log_record = IterationLogRecord(
                iteration=k,
                objective=value(master_data.master_model.epigraph_obj),
                first_stage_var_shift=fsv_shift,
                second_stage_var_shift=ssv_shift,
                dr_var_shift=dr_var_shift,
                num_violated_cons=None,
                max_violation=None,
                dr_polishing_success=polishing_successful,
                all_sep_problems_solved=None,
                global_separation=None,
                elapsed_time=model_data.timing.get_main_elapsed_time(),
                master_backup_solver=master_soln.backup_solver_used,
                master_feasibility_success=master_soln.feasibility_problem_success,
                separation_backup_local_solver=None,
                separation_backup_global_solver=None,
            )
            iter_log_record.log(config.progress_logger.info)
            return GRCSResults(
                master_results=master_soln,
                separation_results=None,
                pyros_termination_condition=pyrosTerminationCondition.time_out,
                iterations=k + 1,
            )

        # === Solve Separation Problem
        separation_data.iteration = k
        separation_data.master_model = master_data.master_model
        separation_results = separation_data.solve_separation(master_data)

        scaled_violations = [
            solve_call_res.scaled_violations[con]
            for con, solve_call_res in separation_results.main_loop_results.solver_call_results.items()
            if solve_call_res.scaled_violations is not None
        ]
        if scaled_violations:
            max_sep_con_violation = max(scaled_violations)
        else:
            max_sep_con_violation = None
        num_violated_cons = len(separation_results.violated_second_stage_ineq_cons)

        all_sep_problems_solved = (
            len(scaled_violations) == num_second_stage_ineq_cons
            and not separation_results.subsolver_error
            and not separation_results.time_out
        ) or separation_results.all_discrete_scenarios_exhausted

        iter_log_record = IterationLogRecord(
            iteration=k,
            objective=value(master_data.master_model.epigraph_obj),
            first_stage_var_shift=fsv_shift,
            second_stage_var_shift=ssv_shift,
            dr_var_shift=dr_var_shift,
            num_violated_cons=num_violated_cons,
            max_violation=max_sep_con_violation,
            dr_polishing_success=polishing_successful,
            all_sep_problems_solved=all_sep_problems_solved,
            global_separation=separation_results.solved_globally,
            elapsed_time=get_main_elapsed_time(model_data.timing),
            master_backup_solver=master_soln.backup_solver_used,
            master_feasibility_success=master_soln.feasibility_problem_success,
            separation_backup_local_solver=separation_results.backup_local_solver_used,
            separation_backup_global_solver=(
                separation_results.backup_global_solver_used
            ),
        )

        # terminate on time limit
        if separation_results.time_out or separation_results.subsolver_error:
            # report PyROS failure to find violated constraint for subsolver error
            if separation_results.subsolver_error:
                config.progress_logger.warning(
                    "PyROS failed to find a constraint violation and "
                    "will terminate with sub-solver error."
                )

            pyros_term_cond = (
                pyrosTerminationCondition.time_out
                if separation_results.time_out
                else pyrosTerminationCondition.subsolver_error
            )
            iter_log_record.log(config.progress_logger.info)
            return GRCSResults(
                master_results=master_soln,
                separation_results=separation_results,
                pyros_termination_condition=pyros_term_cond,
                iterations=k + 1,
            )

        # === Check if we terminate due to robust optimality or feasibility,
        #     or in the event of bypassing global separation, no violations
        robustness_certified = separation_results.robustness_certified
        if robustness_certified:
            if config.bypass_global_separation:
                config.progress_logger.warning(
                    "Option to bypass global separation was chosen. "
                    "Robust feasibility and optimality of the reported "
                    "solution are not guaranteed."
                )
            robust_optimal = (
                config.solve_master_globally
                and config.objective_focus is ObjectiveType.worst_case
            )
            if robust_optimal:
                termination_condition = pyrosTerminationCondition.robust_optimal
            else:
                termination_condition = pyrosTerminationCondition.robust_feasible
            iter_log_record.log(config.progress_logger.info)
            return GRCSResults(
                master_results=master_soln,
                separation_results=separation_results,
                pyros_termination_condition=termination_condition,
                iterations=k + 1,
            )

        # === Add block to master at violation
        mp_methods.add_scenario_block_to_master_problem(
            master_model=master_data.master_model,
            scenario_idx=(k + 1, 0),
            param_realization=separation_results.violating_param_realization,
            from_block=nominal_master_blk,
            clone_first_stage_components=False,
        )

        separation_data.points_added_to_master[(k + 1, 0)] = (
            separation_results.violating_param_realization
        )
        separation_data.auxiliary_values_for_master_points[(k + 1, 0)] = (
            separation_results.auxiliary_param_values
        )

        config.progress_logger.debug("Points added to master:")
        config.progress_logger.debug(
            np.array([pt for pt in separation_data.points_added_to_master.values()])
        )

        # initialize second-stage and state variables
        # for new master block to separation
        # solution chosen by heuristic. consequently,
        # equality constraints should all be satisfied (up to tolerances).
        for var, val in separation_results.violating_separation_variable_values.items():
            master_var = master_data.master_model.scenarios[k + 1, 0].find_component(
                var
            )
            master_var.set_value(val)

        k += 1

        iter_log_record.log(config.progress_logger.info)
        previous_iter_var_data = current_iter_var_data

    # Iteration limit reached
    return GRCSResults(
        master_results=master_soln,
        separation_results=separation_results,
        pyros_termination_condition=pyrosTerminationCondition.max_iter,
        iterations=k,  # iteration count was already incremented
    )



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

"""
This module contains functions for computing an irreducible infeasible set
for a Pyomo MILP or LP using a specified commercial solver, one of CPLEX,
Gurobi, or Xpress.
"""

import abc
import logging
from pyomo.environ import SolverFactory

logger = logging.getLogger("pyomo.contrib.iis")
logger.setLevel(logging.INFO)


def write_iis(pyomo_model, iis_file_name, solver=None, logger=logger):
    """
    Write an irreducible infeasible set for a Pyomo MILP or LP
    using the specified commercial solver.

    Arguments
    ---------
        pyomo_model:
            A Pyomo Block or ConcreteModel
        iis_file_name:str
            A file name to write the IIS to, e.g., infeasible_model.ilp
        solver:str
            Specify the solver to use, one of "cplex", "gurobi", or "xpress".
            If None, the tool will use the first solver available.
        logger:logging.Logger
            A logger for messages. Uses pyomo.contrib.iis logger by default.

    Returns
    -------
        iis_file_name:str
            The file containing the IIS.
    """

    available_solvers = [
        s
        for s, sp in zip(_supported_solvers, _supported_solvers_persistent)
        if SolverFactory(sp).available(exception_flag=False)
    ]

    if solver is None:
        if len(available_solvers) == 0:
            raise RuntimeError(
                f"Could not find a solver to use, supported solvers are {_supported_solvers}"
            )
        solver = available_solvers[0]
        logger.info(f"Using solver {solver}")
    else:
        # validate
        solver = solver.lower()
        solver = _remove_suffix(solver, "_persistent")
        if solver not in available_solvers:
            raise RuntimeError(
                f"The Pyomo persistent interface to {solver} could not be found."
            )

    solver_name = solver
    solver = SolverFactory(solver + "_persistent")

    solver.set_instance(pyomo_model, symbolic_solver_labels=True)

    iis = IISFactory(solver)
    iis.compute()
    iis_file_name = iis.write(iis_file_name)
    logger.info(f"IIS written to {iis_file_name}")
    return iis_file_name


def _remove_suffix(string, suffix):
    if string.endswith(suffix):
        return string[: -len(suffix)]
    else:
        return string


class _IISBase(abc.ABC):
    def __init__(self, solver):
        self._solver = solver

    @abc.abstractmethod
    def compute(self):
        """computes the IIS/Conflict"""
        pass

    @abc.abstractmethod
    def write(self, file_name):
        """writes the IIS in LP format
        return the file name written
        """
        pass


class CplexConflict(_IISBase):
    def compute(self):
        self._solver._solver_model.conflict.refine()

    def write(self, file_name):
        self._solver._solver_model.conflict.write(file_name)
        return file_name


class GurobiIIS(_IISBase):
    def compute(self):
        self._solver._solver_model.computeIIS()

    def write(self, file_name):
        # gurobi relies on the suffix to
        # determine the file type
        file_name = _remove_suffix(file_name, ".ilp")
        file_name += ".ilp"
        self._solver._solver_model.write(file_name)
        return file_name


class XpressIIS(_IISBase):
    def compute(self):
        self._solver._solver_model.iisfirst(1)

    def write(self, file_name):
        self._solver._solver_model.iiswrite(0, file_name, 0, "l")
        if self._solver._version[0] < 38:
            return file_name
        else:
            return _remove_suffix(file_name, ".lp") + ".lp"


_solver_map = {
    "cplex_persistent": CplexConflict,
    "gurobi_persistent": GurobiIIS,
    "xpress_persistent": XpressIIS,
}


def IISFactory(solver):
    if solver.name not in _solver_map:
        raise RuntimeError(f"Unrecognized solver {solver.name}")
    return _solver_map[solver.name](solver)


_supported_solvers_persistent = list(_solver_map.keys())
_supported_solvers = [_remove_suffix(s, "_persistent") for s in _solver_map]



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________
#
#

"""Pyomo Units Container Module

This module provides support for including units within Pyomo expressions. This module
can be used to define units on a model, and to check the consistency of units
within the underlying constraints and expressions in the model. The module also
supports conversion of units within expressions using the `convert` method to support
construction of constraints that contain embedded unit conversions.

To use this package within your Pyomo model, you first need an instance of a
PyomoUnitsContainer. You can use the module level instance already defined as
'units'. This object 'contains' the units - that is, you can access units on
this module using common notation.

    .. doctest::
       :skipif: not pint_available

       >>> from pyomo.environ import units as u
       >>> print(3.0*u.kg)
       3.0*kg

Units can be assigned to Var, Param, and ExternalFunction components, and can
be used directly in expressions (e.g., defining constraints). You can also
verify that the units are consistent on a model, or on individual components
like the objective function, constraint, or expression using
`assert_units_consistent` (from pyomo.util.check_units).
There are other methods there that may be helpful for verifying correct units on a model.

    .. doctest::
       :skipif: not pint_available

       >>> from pyomo.environ import ConcreteModel, Var, Objective
       >>> from pyomo.environ import units as u
       >>> from pyomo.util.check_units import assert_units_consistent, assert_units_equivalent, check_units_equivalent
       >>> model = ConcreteModel()
       >>> model.acc = Var(initialize=5.0, units=u.m/u.s**2)
       >>> model.obj = Objective(expr=(model.acc - 9.81*u.m/u.s**2)**2)
       >>> assert_units_consistent(model.obj) # raise exc if units invalid on obj
       >>> assert_units_consistent(model) # raise exc if units invalid anywhere on the model
       >>> assert_units_equivalent(model.obj.expr, u.m**2/u.s**4) # raise exc if units not equivalent
       >>> print(u.get_units(model.obj.expr)) # print the units on the objective
       m**2/s**4
       >>> print(check_units_equivalent(model.acc, u.m/u.s**2))
       True

The implementation is currently based on the `pint
<http://pint.readthedocs.io>`_ package and supports all the units that
are supported by pint.  The list of units that are supported by pint
can be found at the following url:
https://github.com/hgrecco/pint/blob/master/pint/default_en.txt.

If you need a unit that is not in the standard set of defined units,
you can create your own units by adding to the unit definitions within
pint. See :py:meth:`PyomoUnitsContainer.load_definitions_from_file` or
:py:meth:`PyomoUnitsContainer.load_definitions_from_strings` for more
information.

.. note:: In this implementation of units, "offset" units for
          temperature are not supported within expressions (i.e. the
          non-absolute temperature units including degrees C and
          degrees F).  This is because there are many non-obvious
          combinations that are not allowable. This concern becomes
          clear if you first convert the non-absolute temperature
          units to absolute and then perform the operation. For
          example, if you write 30 degC + 30 degC == 60 degC, but
          convert each entry to Kelvin, the expression is not true
          (i.e., 303.15 K + 303.15 K is not equal to 333.15
          K). Therefore, there are several operations that are not
          allowable with non-absolute units, including addition,
          multiplication, and division.

          This module does support conversion of offset units to
          absolute units numerically, using convert_value_K_to_C,
          convert_value_C_to_K, convert_value_R_to_F,
          convert_value_F_to_R.  These are useful for converting input
          data to absolute units, and for converting data to
          convenient units for reporting.

          Please see the pint documentation `here
          <https://pint.readthedocs.io/en/0.9/nonmult.html>`_ for more
          discussion. While pint implements "delta" units (e.g.,
          delta_degC) to support correct unit conversions, it can be
          difficult to identify and guarantee valid operations in a
          general algebraic modeling environment. While future work
          may support units with relative scale, the current
          implementation requires use of absolute temperature units
          (i.e. K and R) within expressions and a direct conversion of
          numeric values using specific functions for converting input
          data and reporting.

"""

# TODO
#    * create a new pint unit definition file (and load from that file)
#       since the precision in pint seems insufficient for 1e-8 constraint tolerances
#    * Investigate when we can and cannot handle offset units and expand capabilities if possible
#    * Further investigate issues surrounding absolute and relative temperatures (delta units)
#    * Extend external function interface to support units for the arguments in addition to the function itself

import logging
import sys

from pyomo.common.dependencies import pint as pint_module, pint_available
from pyomo.common.modeling import NOTSET
from pyomo.core.expr.expr_common import _type_check_exception_arg
from pyomo.core.expr.numvalue import (
    NumericValue,
    nonpyomo_leaf_types,
    value,
    native_types,
    native_numeric_types,
)
from pyomo.core.expr.template_expr import IndexTemplate
from pyomo.core.expr.visitor import ExpressionValueVisitor
import pyomo.core.expr as EXPR

logger = logging.getLogger(__name__)


class UnitsError(Exception):
    """
    An exception class for all general errors/warnings associated with units
    """

    def __init__(self, msg):
        self.msg = msg

    def __str__(self):
        return str(self.msg)


class InconsistentUnitsError(UnitsError):
    """
    An exception indicating that inconsistent units are present on an expression.

    E.g., x == y, where x is in units of kg and y is in units of meter
    """

    def __init__(self, exp1, exp2, msg):
        msg = f'{msg}: {exp1} not compatible with {exp2}.'
        super(InconsistentUnitsError, self).__init__(msg)


def _pint_unit_mapper(encode, val):
    if encode:
        return str(val)
    else:
        return units._pint_registry(val).units


def _pint_registry_mapper(encode, val):
    if encode:
        if val is not units._pint_registry:
            # FIXME: we currently will not correctly unpickle units
            # associated with a unit manager other than the default
            # singleton.  If we wanted to support this, we would need to
            # do something like create a global units manager registry
            # that would associate each unit manager with a name.  We
            # could then pickle that name and then attempt to restore
            # the association with the original units manager.  As we
            # expect all users to just use the global default, for the
            # time being we will just issue a warning that things may
            # break.
            logger.warning(
                "pickling a _PyomoUnit associated with a PyomoUnitsContainer "
                "that is not the default singleton (%s.units).  Restoring "
                "this state will attempt to return a unit associated with "
                "the default singleton." % (__name__,)
            )
        return None
    elif val is None:
        return units._pint_registry
    else:
        return val


class _PyomoUnit(NumericValue):
    """An object that represents a single unit in Pyomo (e.g., kg, meter)

    Users should not create instances of _PyomoUnit directly, but rather access
    units as attributes on an instance of a :class:`PyomoUnitsContainer`.
    This module contains a global PyomoUnitsContainer object :py:data:`units`.
    See module documentation for more information.
    """

    __slots__ = ('_pint_unit', '_pint_registry')
    __autoslot_mappers__ = {
        '_pint_unit': _pint_unit_mapper,
        '_pint_registry': _pint_registry_mapper,
    }

    def __init__(self, pint_unit, pint_registry):
        super(_PyomoUnit, self).__init__()
        assert pint_unit is not None
        assert pint_registry is not None
        self._pint_unit = pint_unit
        self._pint_registry = pint_registry

    def _get_pint_unit(self):
        """Return the pint unit corresponding to this Pyomo unit."""
        return self._pint_unit

    def _get_pint_registry(self):
        """Return the pint registry (pint.UnitRegistry) object used to create this unit."""
        return self._pint_registry

    def getname(self, fully_qualified=False, name_buffer=None):
        """
        Returns the name of this unit as a string.
        Overloaded from: :py:class:`NumericValue`. See this class for a description of the
        arguments. The value of these arguments are ignored here.

        Returns
        -------
        : str
           Returns the name of the unit
        """
        return str(self)

    # methods/properties that use the NumericValue base class implementation
    # name property
    # local_name
    # cname

    def is_constant(self):
        """
        Indicates if the NumericValue is constant and can be replaced with a plain old number
        Overloaded from: :py:class:`NumericValue`

        This method indicates if the NumericValue is a constant and can be replaced with a plain
        old number. Although units are, in fact, constant, we do NOT want this replaced - therefore
        we return False here to prevent replacement.

        Returns
        =======
        : bool
           False (This method always returns False)
        """
        return False

    def is_fixed(self):
        """
        Indicates if the NumericValue is fixed with respect to a "solver".
        Overloaded from: :py:class:`NumericValue`

        Indicates if the Unit should be treated as fixed. Since the Unit is always treated as
        a constant value of 1.0, it is fixed.

        Returns
        =======
        : bool
           True (This method always returns True)

        """
        return True

    def is_parameter_type(self):
        """This is not a parameter type (overloaded from NumericValue)"""
        return False

    def is_variable_type(self):
        """This is not a variable type (overloaded from NumericValue)"""
        return False

    def is_potentially_variable(self):
        """
        This is not potentially variable (does not and cannot contain a variable).
        Overloaded from NumericValue
        """
        return False

    def is_named_expression_type(self):
        """This is not a named expression (overloaded from NumericValue)"""
        return False

    def is_expression_type(self, expression_system=None):
        """This is a leaf, not an expression (overloaded from NumericValue)"""
        return False

    def is_component_type(self):
        """This is not a component type (overloaded from NumericValue)"""
        return False

    def is_indexed(self):
        """This is not indexed (overloaded from NumericValue)"""
        return False

    def _compute_polynomial_degree(self, result):
        """Returns the polynomial degree - since units are constants, they have degree of zero.
        Note that :py:meth:`NumericValue.polynomial_degree` calls this method.
        """
        return 0

    def __deepcopy__(self, memo):
        # Note that while it is possible to deepcopy the _pint_unit and
        # _pint_registry object (in pint>0.10), that version does not
        # support all Python versions currently supported by Pyomo.
        # Further, Pyomo's use of units relies on a model using a single
        # instance of the pint unit registry.  As we regularly assemble
        # block models using multiple clones (deepcopies) of a base
        # model, it is important that we treat _PyomoUnit objects
        # as outside the model scope and DO NOT duplicate them.
        return self

    def __eq__(self, other):
        if other.__class__ is _PyomoUnit:
            return (
                self._pint_registry is other._pint_registry
                and self._pint_unit == other._pint_unit
            )
        return super().__eq__(other)

    # __bool__ uses NumericValue base class implementation
    # __float__ uses NumericValue base class implementation
    # __int__ uses NumericValue base class implementation
    # __lt__ uses NumericValue base class implementation
    # __gt__ uses NumericValue base class implementation
    # __le__ uses NumericValue base class implementation
    # __ge__ uses NumericValue base class implementation
    # __eq__ uses NumericValue base class implementation
    # __add__ uses NumericValue base class implementation
    # __sub__ uses NumericValue base class implementation
    # __mul__ uses NumericValue base class implementation
    # __div__ uses NumericValue base class implementation
    # __truediv__ uses NumericValue base class implementation
    # __pow__ uses NumericValue vase class implementation
    # __radd__ uses NumericValue base class implementation
    # __rsub__ uses NumericValue base class implementation
    # __rmul__ uses NumericValue base class implementation
    # __rdiv__ uses NumericValue base class implementation
    # __rtruediv__ uses NumericValue base class implementation
    # __rpow__ uses NumericValue base class implementation
    # __iadd__ uses NumericValue base class implementation
    # __isub__ uses NumericValue base class implementation
    # __imul__ uses NumericValue base class implementation
    # __idiv__ uses NumericValue base class implementation
    # __itruediv__ uses NumericValue base class implementation
    # __ipow__ uses NumericValue base class implementation
    # __neg__ uses NumericValue base class implementation
    # __pos__ uses NumericValue base class implementation
    # __add__ uses NumericValue base class implementation

    def __str__(self):
        """Returns a string representing the unit"""

        # The ~ returns the short form of the pint unit if the unit is
        # an instance of the unit 'dimensionless', then pint returns ''
        # which causes problems with some string processing in Pyomo
        # that expects a name
        #
        # Note: Some pint units contain unicode characters (notably
        # delta temperatures).  So that things work cleanly in Python 2
        # and 3, we will generate the string as unicode, then explicitly
        # encode it to UTF-8 in Python 2
        retstr = u'{:~C}'.format(self._pint_unit)
        if retstr == '':
            retstr = 'dimensionless'
        return retstr

    def to_string(self, verbose=None, labeler=None, smap=None, compute_values=False):
        """
        Return a string representation of the expression tree.

        See documentation on :py:class:`NumericValue`

        Returns
        -------
        : bool
           A string representation for the expression tree.
        """
        _str = str(self)
        if any(map(_str.__contains__, ' */')):
            return "(" + _str + ")"
        else:
            return _str

    def __call__(self, exception=NOTSET):
        """Unit is treated as a constant value, and this method always returns 1.0

        Returns
        -------
        : float
           Returns 1.0
        """
        _type_check_exception_arg(self, exception)
        return 1.0

    @property
    def value(self):
        return 1.0

    def pprint(self, ostream=None, verbose=False):
        """Display a user readable string description of this object."""
        if ostream is None:  # pragma:nocover
            ostream = sys.stdout
        ostream.write(str(self))
        # There is also a long form, but the verbose flag is not really the correct indicator
        # if verbose:
        #     ostream.write('{:s}'.format(self._pint_unit))
        # else:
        #     ostream.write('{:~s}'.format(self._pint_unit))


class PintUnitExtractionVisitor(EXPR.StreamBasedExpressionVisitor):
    def __init__(self, pyomo_units_container, units_equivalence_tolerance=1e-12):
        """
        Visitor class used to determine units of an expression. Do not use
        this class directly, but rather use
        "py:meth:`PyomoUnitsContainer.assert_units_consistent`
        or :py:meth:`PyomoUnitsContainer.get_units`

        Parameters
        ----------
        pyomo_units_container : PyomoUnitsContainer
           Instance of the PyomoUnitsContainer that was used for the units
           in the expressions. Pyomo does not support "mixing" units from
           different containers

        units_equivalence_tolerance : float (default 1e-12)
            Floating point tolerance used when deciding if units are equivalent
            or not.

        Notes
        -----
        This class inherits from the :class:`StreamBasedExpressionVisitor` to implement
        a walker that returns the pyomo units and pint units corresponding to an
        expression.

        There are class attributes (dicts) that map the expression node type to the
        particular method that should be called to return the units of the node based
        on the units of its child arguments. This map is used in exitNode.
        """
        super(PintUnitExtractionVisitor, self).__init__()
        self._pyomo_units_container = pyomo_units_container
        self._pint_dimensionless = None
        self._equivalent_pint_units = pyomo_units_container._equivalent_pint_units
        self._equivalent_to_dimensionless = (
            pyomo_units_container._equivalent_to_dimensionless
        )

    def _get_unit_for_equivalent_children(self, node, child_units):
        """
        Return (and test) the units corresponding to an expression node in the
        expression tree where all children should have the same units (e.g. sum).

        Parameters
        ----------
        node : Pyomo expression node
            The parent node of the children

        child_units : list
           This is a list of pint units (one for each of the children)

        Returns
        -------
        : pint unit
        """
        # TODO: This may be expensive for long summations and, in the
        # case of reporting only, we may want to skip the checks
        assert bool(child_units)

        # verify that the pint units are equivalent from each
        # of the child nodes - assume that PyomoUnits are equivalent
        pint_unit_0 = child_units[0]
        for pint_unit_i in child_units:
            if not self._equivalent_pint_units(pint_unit_0, pint_unit_i):
                raise InconsistentUnitsError(
                    pint_unit_0,
                    pint_unit_i,
                    'Error in units found in expression: %s' % (node,),
                )

        # checks were OK, return the first one in the list
        return pint_unit_0

    def _get_unit_for_product(self, node, child_units):
        """
        Return (and test) the units corresponding to a product expression node
        in the expression tree.

        Parameters
        ----------
        node : Pyomo expression node
            The parent node of the children

        child_units : list
           This is a list of pint units (one for each of the children)

        Returns
        -------
        : pint unit
        """
        assert len(child_units) == 2

        pint_unit = child_units[0] * child_units[1]

        if hasattr(pint_unit, 'units'):
            return pint_unit.units

        return pint_unit

    def _get_unit_for_division(self, node, child_units):
        """
        Return (and test) the units corresponding to a division expression node
        in the expression tree.

        Parameters
        ----------
        node : Pyomo expression node
            The parent node of the children

        child_units : list
           This is a list of pint units (one for each of the children)

        Returns
        -------
        : pint unit
        """
        assert len(child_units) == 2

        # this operation can create a quantity, but we want a pint unit object
        return child_units[0] / child_units[1]

    def _get_unit_for_pow(self, node, child_units):
        """
        Return (and test) the units corresponding to a pow expression node
        in the expression tree.

        Parameters
        ----------
        node : Pyomo expression node
            The parent node of the children

        child_units : list
           This is a list of pint units (one for each of the children)

        Returns
        -------
        : pint unit
        """
        assert len(child_units) == 2

        # the exponent needs to be dimensionless
        if not self._equivalent_to_dimensionless(child_units[1]):
            # todo: allow radians?
            raise UnitsError(
                f"Error in sub-expression: {node}. "
                "Exponents in a pow expression must be dimensionless."
            )

        # common case - exponent is a constant number
        exponent = node.args[1]
        if type(exponent) in nonpyomo_leaf_types:
            return child_units[0] ** value(exponent)

        # if base is dimensioness, exponent doesn't matter
        if self._equivalent_to_dimensionless(child_units[0]):
            return self._pint_dimensionless

        # base is not dimensionless, exponent is dimensionless
        # ensure that the exponent is fixed
        if not exponent.is_fixed():
            raise UnitsError(
                f"The base of an exponent has units {child_units[0]}, but "
                "the exponent is not a fixed numerical value."
            )

        return child_units[0] ** value(exponent)

    def _get_unit_for_single_child(self, node, child_units):
        """
        Return (and test) the units corresponding to a unary operation (e.g. negation)
        expression node in the expression tree.

        Parameters
        ----------
        node : Pyomo expression node
            The parent node of the children

        child_units : list
           This is a list of pint units (one for each of the children)

        Returns
        -------
        : pint unit
        """
        assert len(child_units) == 1
        return child_units[0]

    def _get_units_ExternalFunction(self, node, child_units):
        """
        Check to make sure that any child arguments are consistent with
        arg_units return the value from node.get_units() This
        was written for ExternalFunctionExpression where the external
        function has units assigned to its return value and arguments

        Parameters
        ----------
        node : Pyomo expression node
            The parent node of the children

        child_units : list
           This is a list of pint units (one for each of the children)

        Returns
        -------
        : pint unit
        """
        # get the list of arg_units
        arg_units = node.get_arg_units()
        dless = self._pint_dimensionless
        if arg_units is None:
            # they should all be dimensionless
            arg_units = [dless] * len(child_units)
        else:
            # copy arg_units so we don't overwrite the ones in the expression object
            arg_units = list(arg_units)
            for i, a in enumerate(arg_units):
                arg_units[i] = self._pyomo_units_container._get_pint_units(a)

        for arg_unit, pint_unit in zip(arg_units, child_units):
            assert arg_unit is not None
            if not self._equivalent_pint_units(arg_unit, pint_unit):
                raise InconsistentUnitsError(
                    arg_unit, pint_unit, 'Inconsistent units found in ExternalFunction.'
                )

        # now return the units in node.get_units
        return self._pyomo_units_container._get_pint_units(node.get_units())

    def _get_dimensionless_with_dimensionless_children(self, node, child_units):
        """
        Check to make sure that any child arguments are unitless /
        dimensionless (for functions like exp()) and return dimensionless.

        Parameters
        ----------
        node : Pyomo expression node
            The parent node of the children

        child_units : list
           This is a list of pint units (one for each of the children)

        Returns
        -------
        : pint unit
        """
        for pint_unit in child_units:
            if not self._equivalent_to_dimensionless(pint_unit):
                raise UnitsError(
                    f'Expected no units or dimensionless units in {node}, '
                    f'but found {pint_unit}.'
                )

        return self._pint_dimensionless

    def _get_dimensionless_no_children(self, node, child_units):
        """
        Check to make sure the length of child_units is zero, and returns
        dimensionless. Used for leaf nodes that should not have any units.

        Parameters
        ----------
        node : Pyomo expression node
            The parent node of the children

        child_units : list
           This is a list of pint units (one for each of the children)

        Returns
        -------
        : pint unit
        """
        assert len(child_units) == 0
        # may need more checks for dimensionless for other types
        assert type(node) is IndexTemplate
        return self._pint_dimensionless

    def _get_unit_for_unary_function(self, node, child_units):
        """
        Return (and test) the units corresponding to a unary function expression node
        in the expression tree. Checks that child_units is of length 1
        and calls the appropriate method from the unary function method map.

        Parameters
        ----------
        node : Pyomo expression node
            The parent node of the children

        child_units : list
           This is a list of pint units (one for each of the children)

        Returns
        -------
        : pint unit
        """
        assert len(child_units) == 1
        func_name = node.getname()
        node_func = self.unary_function_method_map.get(func_name, None)
        if node_func is None:
            raise TypeError(
                f'An unhandled unary function: {func_name} was encountered '
                f'while retrieving the units of expression {node}'
            )
        return node_func(self, node, child_units)

    def _get_unit_for_expr_if(self, node, child_units):
        """
        Return (and test) the units corresponding to an Expr_if expression node
        in the expression tree. The _if relational expression is validated and
        the _then/_else are checked to ensure they have the same units. Also checks
        to make sure length of child_units is 3

        Parameters
        ----------
        node : Pyomo expression node
            The parent node of the children

        child_units : list
           This is a list of pint units (one for each of the children)

        Returns
        -------
        : pint unit
        """
        assert len(child_units) == 3

        # the _if should already be consistent (since the children were
        # already checked)
        if not self._equivalent_pint_units(child_units[1], child_units[2]):
            raise InconsistentUnitsError(
                child_units[1],
                child_units[2],
                'Error in units found in expression: %s' % (node,),
            )

        return child_units[1]

    def _get_dimensionless_with_radians_child(self, node, child_units):
        """
        Return (and test) the units corresponding to a trig function expression node
        in the expression tree. Checks that the length of child_units is 1
        and that the units of that child expression are unitless or radians and
        returns dimensionless for the units.

        Parameters
        ----------
        node : Pyomo expression node
            The parent node of the children

        child_units : list
           This is a list of pint units (one for each of the children)

        Returns
        -------
        : pint unit
        """
        assert len(child_units) == 1

        if self._equivalent_to_dimensionless(child_units[0]):
            return self._pint_dimensionless
        if self._equivalent_pint_units(
            child_units[0], self._pyomo_units_container._pint_registry.radian
        ):
            return self._pint_dimensionless

        # units are not None, dimensionless, or radians
        raise UnitsError(
            'Expected radians or dimensionless in argument to function '
            'in expression %s, but found %s' % (node, child_units[0])
        )

    def _get_radians_with_dimensionless_child(self, node, child_units):
        """
        Return (and test) the units corresponding to an inverse trig expression node
        in the expression tree. Checks that the length of child_units is 1
        and that the child argument is dimensionless, and returns radians

        Parameters
        ----------
        node : Pyomo expression node
            The parent node of the children

        child_units : list
           This is a list of pint units (one for each of the children)

        Returns
        -------
        : pint unit
        """
        assert len(child_units) == 1

        if self._equivalent_to_dimensionless(child_units[0]):
            return self._pyomo_units_container._pint_registry.radian

        raise UnitsError(
            f'Expected dimensionless argument to function in expression {node},'
            f' but found {child_units[0]}'
        )

    def _get_unit_sqrt(self, node, child_units):
        """
        Return (and test) the units corresponding to a sqrt expression node
        in the expression tree. Checks that the length of child_units is one.

        Parameters
        ----------
        node : Pyomo expression node
            The parent node of the children

        child_units : list
           This is a list of pint units (one for each of the children)

        Returns
        -------
        : pint unit
        """
        assert len(child_units) == 1
        return child_units[0] ** 0.5

    node_type_method_map = {
        EXPR.EqualityExpression: _get_unit_for_equivalent_children,
        EXPR.InequalityExpression: _get_unit_for_equivalent_children,
        EXPR.RangedExpression: _get_unit_for_equivalent_children,
        EXPR.SumExpression: _get_unit_for_equivalent_children,
        EXPR.NPV_SumExpression: _get_unit_for_equivalent_children,
        EXPR.ProductExpression: _get_unit_for_product,
        EXPR.MonomialTermExpression: _get_unit_for_product,
        EXPR.NPV_ProductExpression: _get_unit_for_product,
        EXPR.DivisionExpression: _get_unit_for_division,
        EXPR.NPV_DivisionExpression: _get_unit_for_division,
        EXPR.PowExpression: _get_unit_for_pow,
        EXPR.NPV_PowExpression: _get_unit_for_pow,
        EXPR.NegationExpression: _get_unit_for_single_child,
        EXPR.NPV_NegationExpression: _get_unit_for_single_child,
        EXPR.AbsExpression: _get_unit_for_single_child,
        EXPR.NPV_AbsExpression: _get_unit_for_single_child,
        EXPR.UnaryFunctionExpression: _get_unit_for_unary_function,
        EXPR.NPV_UnaryFunctionExpression: _get_unit_for_unary_function,
        EXPR.Expr_ifExpression: _get_unit_for_expr_if,
        IndexTemplate: _get_dimensionless_no_children,
        EXPR.Numeric_GetItemExpression: (
            _get_dimensionless_with_dimensionless_children
        ),
        EXPR.NPV_Numeric_GetItemExpression: (
            _get_dimensionless_with_dimensionless_children
        ),
        EXPR.ExternalFunctionExpression: _get_units_ExternalFunction,
        EXPR.NPV_ExternalFunctionExpression: _get_units_ExternalFunction,
        EXPR.LinearExpression: _get_unit_for_equivalent_children,
    }

    unary_function_method_map = {
        'log': _get_dimensionless_with_dimensionless_children,
        'log10': _get_dimensionless_with_dimensionless_children,
        'sin': _get_dimensionless_with_radians_child,
        'cos': _get_dimensionless_with_radians_child,
        'tan': _get_dimensionless_with_radians_child,
        'sinh': _get_dimensionless_with_radians_child,
        'cosh': _get_dimensionless_with_radians_child,
        'tanh': _get_dimensionless_with_radians_child,
        'asin': _get_radians_with_dimensionless_child,
        'acos': _get_radians_with_dimensionless_child,
        'atan': _get_radians_with_dimensionless_child,
        'exp': _get_dimensionless_with_dimensionless_children,
        'sqrt': _get_unit_sqrt,
        'asinh': _get_radians_with_dimensionless_child,
        'acosh': _get_radians_with_dimensionless_child,
        'atanh': _get_radians_with_dimensionless_child,
        'ceil': _get_unit_for_single_child,
        'floor': _get_unit_for_single_child,
    }

    def initializeWalker(self, expr):
        # Refresh the cached dimensionless (in case the underlying pint
        # registry was either changed or had not been set when the
        # PyomoUnitsContainer was originally created).
        self._pint_dimensionless = self._pyomo_units_container._pint_dimensionless
        walk, result = self.beforeChild(None, expr, 0)
        if not walk:
            result = self.finalizeResult(result)
        return walk, result

    def beforeChild(self, node, child, child_idx):
        ctype = child.__class__
        if ctype in native_types:
            return False, self._pint_dimensionless

        if child.is_expression_type():
            return True, None

        # this is a leaf, but not a native type
        if ctype is _PyomoUnit:
            return False, child._get_pint_unit()
        elif hasattr(child, 'get_units'):
            # might want to add other common types here
            pyomo_unit = child.get_units()
            pint_unit = self._pyomo_units_container._get_pint_units(pyomo_unit)
            return False, pint_unit

        return False, self._pint_dimensionless

    def exitNode(self, node, data):
        """Visitor callback when moving up the expression tree.

        Callback for
        :class:`pyomo.core.current.StreamBasedExpressionVisitor`. This
        method is called when moving back up the tree in a depth first
        search.

        """
        node_func = self.node_type_method_map.get(node.__class__, None)
        if node_func is not None:
            return node_func(self, node, data)

        # not a leaf - check if it is a named expression
        if (
            hasattr(node, 'is_named_expression_type')
            and node.is_named_expression_type()
        ):
            pint_unit = self._get_unit_for_single_child(node, data)
            return pint_unit

        raise TypeError(
            f'An unhandled expression node type: {type(node)} was encountered '
            f'while retrieving the units of expression {node}'
        )

    def finalizeResult(self, result):
        if hasattr(result, 'units'):
            # likely, we got a quantity object and not a units object
            return result.units
        return result


class PyomoUnitsContainer:
    """Class that is used to create and contain units in Pyomo.

    This is the class that is used to create, contain, and interact
    with units in Pyomo.  The module
    (:mod:`pyomo.core.base.units_container`) also contains a module
    level units container :py:data:`units` that is an instance of a
    PyomoUnitsContainer. This module instance should typically be used
    instead of creating your own instance of a
    :py:class:`PyomoUnitsContainer`.  For an overview of the usage of
    this class, see the module documentation
    (:mod:`pyomo.core.base.units_container`)

    This class is based on the "pint" module. Documentation for
    available units can be found at the following url:
    https://github.com/hgrecco/pint/blob/master/pint/default_en.txt

    .. note::

        Pre-defined units can be accessed through attributes on the
        PyomoUnitsContainer class; however, these attributes are created
        dynamically through the __getattr__ method, and are not present
        on the class until they are requested.

    """

    def __init__(self, pint_registry=NOTSET):
        """Create a PyomoUnitsContainer instance."""
        if pint_registry is NOTSET:
            pint_registry = pint_module.UnitRegistry()
        self._pint_registry = pint_registry
        if pint_registry is None:
            self._pint_dimensionless = None
        else:
            self._pint_dimensionless = self._pint_registry.dimensionless
        self._pintUnitExtractionVisitor = PintUnitExtractionVisitor(self)

    def load_definitions_from_file(self, definition_file):
        """Load new units definitions from a file

        This method loads additional units definitions from a user
        specified definition file. An example of a definitions file
        can be found at:
        https://github.com/hgrecco/pint/blob/master/pint/default_en.txt

        If we have a file called ``my_additional_units.txt`` with the
        following lines::

            USD = [currency]

        Then we can add this to the container with:

        .. doctest::
            :skipif: not pint_available
            :hide:

            # Get a local units object (to avoid duplicate registration
            # with the example in load_definitions_from_strings)
            >>> import pyomo.core.base.units_container as _units
            >>> u = _units.PyomoUnitsContainer()
            >>> with open('my_additional_units.txt', 'w') as FILE:
            ...     tmp = FILE.write("USD = [currency]\\n")

        .. doctest::
            :skipif: not pint_available

            >>> u.load_definitions_from_file('my_additional_units.txt')
            >>> print(u.USD)
            USD

        .. doctest::
            :skipif: not pint_available
            :hide:

            # Clean up the file we just created
            >>> import os
            >>> os.remove('my_additional_units.txt')

        """
        self._pint_registry.load_definitions(definition_file)
        self._pint_dimensionless = self._pint_registry.dimensionless

    def load_definitions_from_strings(self, definition_string_list):
        """Load new units definitions from a string

        This method loads additional units definitions from a list of
        strings (one for each line). An example of the definitions
        strings can be found at:
        https://github.com/hgrecco/pint/blob/master/pint/default_en.txt

        For example, to add the currency dimension and US dollars as a
        unit, use

        .. doctest::
            :skipif: not pint_available
            :hide:

            # get a local units object (to avoid duplicate registration
            # with the example in load_definitions_from_strings)
            >>> import pint
            >>> import pyomo.core.base.units_container as _units
            >>> u = _units.PyomoUnitsContainer()

        .. doctest::
            :skipif: not pint_available

            >>> u.load_definitions_from_strings(['USD = [currency]'])
            >>> print(u.USD)
            USD

        """
        self._pint_registry.load_definitions(definition_string_list)

    def __getattr__(self, item):
        """Here, __getattr__ is implemented to automatically create the
        necessary unit if the attribute does not already exist.

        Parameters
        ----------
        item : str
            the name of the new field requested external

        Returns
        -------
        PyomoUnit
            returns a PyomoUnit corresponding to the requested attribute,
            or None if it cannot be created.

        """
        # since __getattr__ was called, we must not have this field yet
        # try to build a unit from the requested item
        pint_registry = self._pint_registry
        try:
            pint_unit = getattr(pint_registry, item)

            if pint_unit is not None:
                # check if the unit is an offset unit and throw an exception if necessary
                # TODO: should we prevent delta versions: delta_degC and delta_degF as well?
                pint_unit_container = pint_module.util.to_units_container(
                    pint_unit, pint_registry
                )
                for u, e in pint_unit_container.items():
                    if not pint_registry._units[u].is_multiplicative:
                        raise UnitsError(
                            'Pyomo units system does not support the offset '
                            f'units "{item}". Use absolute units '
                            '(e.g. kelvin instead of degC) instead.'
                        )

                unit = _PyomoUnit(pint_unit, pint_registry)
                setattr(self, item, unit)
                return unit
        except pint_module.errors.UndefinedUnitError as exc:
            pint_unit = None

        if pint_unit is None:
            raise AttributeError(f'Attribute {item} not found.')

    # We added support to specify a units definition file instead of this programmatic interface
    # def create_new_base_dimension(self, dimension_name, base_unit_name):
    #     """
    #     Use this method to create a new base dimension (e.g. a new dimension other than Length, Mass) for the unit manager.
    #
    #     Parameters
    #     ----------
    #     dimension_name : str
    #        name of the new dimension (needs to be unique from other dimension names)
    #
    #     base_unit_name : str
    #        base_unit_name: name of the base unit for this dimension
    #
    #     """
    #     # TODO: Error checking - if dimension already exists then we should return a useful error message.
    #     defn_str = str(base_unit_name) + ' = [' + str(dimension_name) + ']'
    #     self._pint_registry.define(defn_str)
    #
    # def create_new_unit(self, unit_name, base_unit_name, conv_factor, conv_offset=None):
    #     """
    #     Create a new unit that is not already included in the units manager.
    #
    #     Examples:
    #         # create a new unit of length called football field that is 1000 yards
    #         # defines: x (in yards) = y (in football fields) X 100.0
    #         >>> um.create_new_unit('football_field', 'yards', 100.0)
    #
    #         # create a new unit of temperature that is half the size of a degree F
    #         # defines x (in K) = y (in half degF) X 10/9 + 255.3722222 K
    #         >>> um.create_new_unit('half_degF', 'kelvin', 10.0/9.0, 255.3722222)
    #
    #     Parameters
    #     ----------
    #     unit_name : str
    #        name of the new unit to create
    #     base_unit_name : str
    #        name of the base unit from the same "dimension" as the new unit
    #     conv_factor : float
    #        value of the multiplicative factor needed to convert the new unit
    #        to the base unit
    #     conv_offset : float
    #        value of any offset between the new unit and the base unit
    #        Note that the units of this offset are the same as the base unit,
    #        and it is applied after the factor conversion
    #        (e.g., base_value = new_value*conv_factor + conv_offset)
    #
    #     """
    #     if conv_offset is None:
    #         defn_str = '{0!s} = {1:g} * {2!s}'.format(unit_name, float(conv_factor), base_unit_name)
    #     else:
    #         defn_str = '{0!s} = {1:17.16g} * {2!s}; offset: {3:17.16g}'.format(unit_name, float(conv_factor), base_unit_name,
    #                                                                  float(conv_offset))
    #     self._pint_registry.define(defn_str)

    def _rel_diff(self, a, b):
        scale = min(abs(a), abs(b))
        if scale < 1.0:
            scale = 1.0
        return abs(a - b) / scale

    def _equivalent_pint_units(self, a, b, TOL=1e-12):
        if a is b or a == b:
            return True
        base_a = self._pint_registry.get_base_units(a)
        base_b = self._pint_registry.get_base_units(b)
        if base_a[1] != base_b[1]:
            uc_a = base_a[1].dimensionality
            uc_b = base_b[1].dimensionality
            for key in uc_a.keys() | uc_b.keys():
                if self._rel_diff(uc_a.get(key, 0), uc_b.get(key, 0)) >= TOL:
                    return False
        return self._rel_diff(base_a[0], base_b[0]) <= TOL

    def _equivalent_to_dimensionless(self, a, TOL=1e-12):
        if a is self._pint_dimensionless or a == self._pint_dimensionless:
            return True
        base_a = self._pint_registry.get_base_units(a)
        if not base_a[1].dimensionless:
            return False
        return self._rel_diff(base_a[0], 1.0) <= TOL

    def _get_pint_units(self, expr):
        """
        Return the pint units corresponding to the expression. This does
        a number of checks as well.

        Parameters
        ----------
        expr : Pyomo expression
           the input expression for extracting units

        Returns
        -------
        : pint unit
        """
        if expr is None:
            return self._pint_dimensionless
        return self._pintUnitExtractionVisitor.walk_expression(expr=expr)

    def get_units(self, expr):
        """Return the Pyomo units corresponding to this expression (also
        performs validation and will raise an exception if units are not
        consistent).

        Parameters
        ----------
        expr : Pyomo expression
            The expression containing the desired units

        Returns
        -------
        : Pyomo unit (expression)
           Returns the units corresponding to the expression

        Raises
        ------
        :py:class:`pyomo.core.base.units_container.UnitsError`, :py:class:`pyomo.core.base.units_container.InconsistentUnitsError`

        """
        return _PyomoUnit(self._get_pint_units(expr), self._pint_registry)

    def _pint_convert_temp_from_to(
        self, numerical_value, pint_from_units, pint_to_units
    ):
        if type(numerical_value) not in native_numeric_types:
            raise UnitsError(
                'Conversion routines for absolute and relative temperatures '
                'require a numerical value only. Pyomo objects (Var, Param, '
                'expressions) are not supported. Please use value(x) to '
                'extract the numerical value if necessary.'
            )

        src_quantity = self._pint_registry.Quantity(numerical_value, pint_from_units)
        dest_quantity = src_quantity.to(pint_to_units)
        return dest_quantity.magnitude

    def convert_temp_K_to_C(self, value_in_K):
        """
        Convert a value in Kelvin to degrees Celsius.  Note that this method
        converts a numerical value only. If you need temperature
        conversions in expressions, please work in absolute
        temperatures only.
        """
        return self._pint_convert_temp_from_to(
            value_in_K, self._pint_registry.K, self._pint_registry.degC
        )

    def convert_temp_C_to_K(self, value_in_C):
        """
        Convert a value in degrees Celsius to Kelvin Note that this
        method converts a numerical value only. If you need
        temperature conversions in expressions, please work in
        absolute temperatures only.
        """
        return self._pint_convert_temp_from_to(
            value_in_C, self._pint_registry.degC, self._pint_registry.K
        )

    def convert_temp_R_to_F(self, value_in_R):
        """
        Convert a value in Rankine to degrees Fahrenheit.  Note that
        this method converts a numerical value only. If you need
        temperature conversions in expressions, please work in
        absolute temperatures only.
        """
        return self._pint_convert_temp_from_to(
            value_in_R, self._pint_registry.rankine, self._pint_registry.degF
        )

    def convert_temp_F_to_R(self, value_in_F):
        """
        Convert a value in degrees Fahrenheit to Rankine.  Note that
        this method converts a numerical value only. If you need
        temperature conversions in expressions, please work in
        absolute temperatures only.
        """
        return self._pint_convert_temp_from_to(
            value_in_F, self._pint_registry.degF, self._pint_registry.rankine
        )

    def convert(self, src, to_units=None):
        """
        This method returns an expression that contains the
        explicit conversion from one unit to another.

        Parameters
        ----------
        src : Pyomo expression
           The source value that will be converted. This could be a
           Pyomo Var, Pyomo Param, or a more complex expression.
        to_units : Pyomo units expression
           The desired target units for the new expression

        Returns
        -------
           ret : Pyomo expression
        """
        src_pint_unit = self._get_pint_units(src)
        to_pint_unit = self._get_pint_units(to_units)
        if src_pint_unit == to_pint_unit:
            return src

        # We disallow offset units, so we only need a factor to convert
        # between the two
        src_base_factor, base_units_src = self._pint_registry.get_base_units(
            src_pint_unit, check_nonmult=True
        )
        to_base_factor, base_units_to = self._pint_registry.get_base_units(
            to_pint_unit, check_nonmult=True
        )

        if base_units_src != base_units_to:
            raise InconsistentUnitsError(
                src_pint_unit, to_pint_unit, 'Error in convert: units not compatible.'
            )

        return (
            (src_base_factor / to_base_factor)
            * _PyomoUnit(to_pint_unit / src_pint_unit, self._pint_registry)
            * src
        )

    def convert_value(self, num_value, from_units=None, to_units=None):
        """
        This method performs explicit conversion of a numerical value
        from one unit to another, and returns the new value.

        The argument "num_value" must be a native numeric type (e.g. float).
        Note that this method returns a numerical value only, and not an
        expression with units.

        Parameters
        ----------
        num_value : float or other native numeric type
           The value that will be converted
        from_units : Pyomo units expression
           The units to convert from
        to_units : Pyomo units expression
           The units to convert to

        Returns
        -------
           float : The converted value

        """
        if type(num_value) not in native_numeric_types:
            raise UnitsError(
                'The argument "num_value" in convert_value must be a native '
                'numeric type, but instead type {type(num_value)} was found.'
            )

        from_pint_unit = self._get_pint_units(from_units)
        to_pint_unit = self._get_pint_units(to_units)
        if from_pint_unit == to_pint_unit:
            return num_value

        # We disallow offset units, so we only need a factor to convert
        # between the two
        #
        # TODO: Do we need to disallow offset units here? Should we
        # assume the user knows what they are doing?
        #
        # TODO: This check may be overkill - pint will raise an error
        # that may be sufficient
        from_base_factor, from_base_units = self._pint_registry.get_base_units(
            from_pint_unit, check_nonmult=True
        )
        to_base_factor, to_base_units = self._pint_registry.get_base_units(
            to_pint_unit, check_nonmult=True
        )
        if from_base_units != to_base_units:
            raise UnitsError(
                'Cannot convert %s to %s. Units are not compatible.'
                % (from_units, to_units)
            )

        # convert the values
        from_quantity = num_value * from_pint_unit
        to_quantity = from_quantity.to(to_pint_unit)
        return to_quantity.magnitude

    def set_pint_registry(self, pint_registry):
        if pint_registry is self._pint_registry:
            return
        if self._pint_registry is not None:
            logger.warning(
                "Changing the pint registry used by the Pyomo Units "
                "system after the PyomoUnitsContainer was constructed.  "
                "Pint requires that all units and dimensioned quantities "
                "are generated by a single pint registry."
            )
        self._pint_registry = pint_registry
        self._pint_dimensionless = self._pint_registry.dimensionless

    @property
    def pint_registry(self):
        return self._pint_registry


class _QuantityVisitor(ExpressionValueVisitor):
    def __init__(self):
        self.native_types = set(nonpyomo_leaf_types)
        self.native_types.add(units._pint_registry.Quantity)
        self._unary_inverse_trig = {'asin', 'acos', 'atan', 'asinh', 'acosh', 'atanh'}

    def visit(self, node, values):
        """Visit nodes that have been expanded"""
        if node.__class__ in self.handlers:
            return self.handlers[node.__class__](self, node, values)
        return node._apply_operation(values)

    def visiting_potential_leaf(self, node):
        """
        Visiting a potential leaf.

        Return True if the node is not expanded.
        """
        if node.__class__ in self.native_types:
            return True, node

        if node.is_expression_type():
            return False, None

        if node.is_numeric_type():
            if hasattr(node, 'get_units'):
                unit = node.get_units()
                if unit is not None:
                    return True, value(node) * unit._pint_unit
                else:
                    return True, value(node)
            elif node.__class__ is _PyomoUnit:
                return True, node._pint_unit
            else:
                return True, value(node)
        elif node.is_logical_type():
            return True, value(node)
        else:
            return True, node

    def finalize(self, val):
        if val.__class__ is units._pint_registry.Quantity:
            return val
        elif val.__class__ is units._pint_registry.Unit:
            return 1.0 * val
        # else
        try:
            return val * units._pint_dimensionless
        except:
            return val

    def _handle_unary_function(self, node, values):
        ans = node._apply_operation(values)
        if node.getname() in self._unary_inverse_trig:
            ans = ans * units._pint_registry.radian
        return ans

    def _handle_external(self, node, values):
        # External functions are units-unaware
        ans = node._apply_operation(
            [
                val.magnitude if val.__class__ is units._pint_registry.Quantity else val
                for val in values
            ]
        )
        unit = node.get_units()
        if unit is not None:
            ans = ans * unit._pint_unit
        return ans


_QuantityVisitor.handlers = {
    EXPR.UnaryFunctionExpression: _QuantityVisitor._handle_unary_function,
    EXPR.NPV_UnaryFunctionExpression: _QuantityVisitor._handle_unary_function,
    EXPR.ExternalFunctionExpression: _QuantityVisitor._handle_external,
    EXPR.NPV_ExternalFunctionExpression: _QuantityVisitor._handle_external,
}


def as_quantity(expr):
    return _QuantityVisitor().dfs_postorder_stack(expr)


class _DeferredUnitsSingleton(PyomoUnitsContainer):
    """A class supporting deferred interrogation of pint_available.

    This class supports creating a module-level singleton, but deferring
    the interrogation of the pint_available flag until the first time
    the object is actually used.  If pint is available, this instance
    object is replaced by an actual PyomoUnitsContainer.  Otherwise this
    leverages the pint_module to raise an (informative)
    DeferredImportError exception.

    """

    def __init__(self):
        # do NOT call the base class __init__ so that the pint_module is
        # not accessed
        pass

    def __getattribute__(self, attr):
        # Note that this method will only be called ONCE: either pint is
        # present, at which point this instance __class__ will fall back
        # to PyomoUnitsContainer (where this method is not declared, OR
        # pint is not available and an ImportError will be raised.
        #
        # We need special case handling for __class__: gurobipy
        # interrogates things by looking at their __class__ during
        # python shutdown.  Unfortunately, interrogating this
        # singleton's __class__ evaluates `pint_available`, which - if
        # DASK is installed - imports dask.  Importing dask creates
        # threading objects.  Unfortunately, creating threading objects
        # during interpreter shutdown generates a RuntimeError.  So, our
        # solution is to special-case the resolution of __class__ here
        # to avoid accidentally triggering the imports.
        if attr == "__class__":
            return _DeferredUnitsSingleton
        #
        if pint_available:
            # If the first thing that is being called is
            # "units.set_pint_registry(...)", then we will call __init__
            # with None so that the subsequent call to set_pint_registry
            # will work cleanly.  In all other cases, we will initialize
            # PyomoUnitsContainer with a new (default) pint registry.
            if attr == 'set_pint_registry':
                pint_registry = None
            else:
                pint_registry = pint_module.UnitRegistry()
            self.__class__ = PyomoUnitsContainer
            self.__init__(pint_registry)
            return getattr(self, attr)
        else:
            # Generate the ImportError
            return getattr(pint_module, attr)


# Define a module level instance of a PyomoUnitsContainer to use for
# all units within a Pyomo model. If pint is not available, this will
# cause an error at the first usage See module level documentation for
# an example.
units = _DeferredUnitsSingleton()



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

from pyomo.common import unittest

from pyomo.common.dependencies import numpy as numpy, numpy_available

import pyomo.environ as pyo
import pyomo.common.unittest as unittest
from pyomo.common.collections import ComponentSet

import pyomo.contrib.alternative_solutions.aos_utils as au


class TestAOSUtilsUnit(unittest.TestCase):

    def get_multiple_objective_model(self):
        """Create a simple model with three objectives."""
        m = pyo.ConcreteModel()
        m.b1 = pyo.Block()
        m.b2 = pyo.Block()
        m.x = pyo.Var()
        m.y = pyo.Var()
        m.b1.o = pyo.Objective(expr=m.x)
        m.b2.o = pyo.Objective([0, 1])
        m.b2.o[0] = pyo.Objective(expr=m.y)
        m.b2.o[1] = pyo.Objective(expr=m.x + m.y)
        return m

    def test_multiple_objectives(self):
        """Check that an error is thrown with multiple objectives."""
        m = self.get_multiple_objective_model()
        assert_text = (
            "Model has 3 active objective functions, exactly one " "is required."
        )
        with self.assertRaisesRegex(AssertionError, assert_text):
            au.get_active_objective(m)

    def test_no_objectives(self):
        """Check that an error is thrown with no objectives."""
        m = self.get_multiple_objective_model()
        m.b1.o.deactivate()
        m.b2.o.deactivate()
        assert_text = (
            "Model has 0 active objective functions, exactly one " "is required."
        )
        with self.assertRaisesRegex(AssertionError, assert_text):
            au.get_active_objective(m)

    def test_one_objective(self):
        """
        Check that the active objective is returned, when there is just one
        objective.
        """
        m = self.get_multiple_objective_model()
        m.b1.o.deactivate()
        m.b2.o[0].deactivate()
        self.assertEqual(m.b2.o[1], au.get_active_objective(m))

    def test_aos_block(self):
        """Ensure that an alternative solution block is added."""
        m = self.get_multiple_objective_model()
        block_name = "test_block"
        b = au._add_aos_block(m, block_name)
        self.assertEqual(b.name, block_name)
        self.assertEqual(b.ctype, pyo.Block)

    def get_simple_model(self, sense=pyo.minimize):
        """Create a simple 2d linear program with an objective."""
        m = pyo.ConcreteModel()
        m.x = pyo.Var()
        m.y = pyo.Var()
        m.o = pyo.Objective(expr=m.x + m.y, sense=sense)
        return m

    def test_no_obj_constraint(self):
        """Ensure that no objective constraints are added."""
        m = self.get_simple_model()
        cons = au._add_objective_constraint(m, m.o, 2, None, None)
        self.assertEqual(cons, [])
        self.assertEqual(m.find_component("optimality_tol_rel"), None)
        self.assertEqual(m.find_component("optimality_tol_abs"), None)

    def test_min_rel_obj_constraint(self):
        """Ensure that the correct relative objective constraint is added."""
        m = self.get_simple_model()
        cons = au._add_objective_constraint(m, m.o, 2, 0.1, None)
        self.assertEqual(len(cons), 1)
        self.assertEqual(m.find_component("optimality_tol_rel"), cons[0])
        self.assertEqual(m.find_component("optimality_tol_abs"), None)
        self.assertEqual(2.2, cons[0].upper)
        self.assertEqual(None, cons[0].lower)

    def test_min_abs_obj_constraint(self):
        """Ensure that the correct absolute objective constraint is added."""
        m = self.get_simple_model()
        cons = au._add_objective_constraint(m, m.o, 2, None, 1)
        self.assertEqual(len(cons), 1)
        self.assertEqual(m.find_component("optimality_tol_rel"), None)
        self.assertEqual(m.find_component("optimality_tol_abs"), cons[0])
        self.assertEqual(3, cons[0].upper)
        self.assertEqual(None, cons[0].lower)

    def test_min_both_obj_constraint(self):
        m = self.get_simple_model()
        cons = au._add_objective_constraint(m, m.o, -10, 0.3, 5)
        self.assertEqual(len(cons), 2)
        self.assertEqual(m.find_component("optimality_tol_rel"), cons[0])
        self.assertEqual(m.find_component("optimality_tol_abs"), cons[1])
        self.assertEqual(-7, cons[0].upper)
        self.assertEqual(None, cons[0].lower)
        self.assertEqual(-5, cons[1].upper)
        self.assertEqual(None, cons[1].lower)

    def test_max_both_obj_constraint(self):
        """
        Ensure that the correct relative and absolute objective constraints are
        added.
        """
        m = self.get_simple_model(sense=pyo.maximize)
        cons = au._add_objective_constraint(m, m.o, -1, 0.3, 1)
        self.assertEqual(len(cons), 2)
        self.assertEqual(m.find_component("optimality_tol_rel"), cons[0])
        self.assertEqual(m.find_component("optimality_tol_abs"), cons[1])
        self.assertEqual(None, cons[0].upper)
        self.assertEqual(-1.3, cons[0].lower)
        self.assertEqual(None, cons[1].upper)
        self.assertEqual(-2, cons[1].lower)

    def test_max_both_obj_constraint2(self):
        """
        Ensure that the correct relative and absolute objective constraints are
        added.
        """
        m = self.get_simple_model(sense=pyo.maximize)
        cons = au._add_objective_constraint(m, m.o, 20, 0.5, 11)
        self.assertEqual(len(cons), 2)
        self.assertEqual(m.find_component("optimality_tol_rel"), cons[0])
        self.assertEqual(m.find_component("optimality_tol_abs"), cons[1])
        self.assertEqual(None, cons[0].upper)
        self.assertEqual(10, cons[0].lower)
        self.assertEqual(None, cons[1].upper)
        self.assertEqual(9, cons[1].lower)

    @unittest.skipIf(not numpy_available, "Numpy not installed")
    def test_random_direction(self):
        """
        Ensure that _get_random_direction returns a normal vector.
        """
        from numpy.linalg import norm

        vector = au._get_random_direction(10)
        self.assertAlmostEqual(1.0, norm(vector))

    def get_var_model(self):
        """
        Create a model with multiple variables that are nested over several
        layers of blocks.
        """

        indices = [0, 1, 2, 3]

        m = pyo.ConcreteModel()

        m.b1 = pyo.Block()
        m.b2 = pyo.Block()
        m.b1.sb1 = pyo.Block()
        m.b2.sb2 = pyo.Block()

        m.x = pyo.Var(domain=pyo.Reals)
        m.b1.y = pyo.Var(domain=pyo.Binary)
        m.b2.z = pyo.Var(domain=pyo.Integers)

        m.x_f = pyo.Var(domain=pyo.Reals)
        m.b1.y_f = pyo.Var(domain=pyo.Binary)
        m.b2.z_f = pyo.Var(domain=pyo.Integers)
        m.x_f.fix(0)
        m.b1.y_f.fix(0)
        m.b2.z_f.fix(0)

        m.b1.sb1.x_l = pyo.Var(indices, domain=pyo.Reals)
        m.b1.sb1.y_l = pyo.Var(indices, domain=pyo.Binary)
        m.b2.sb2.z_l = pyo.Var(indices, domain=pyo.Integers)

        m.b1.sb1.x_l[3].fix(0)
        m.b1.sb1.y_l[3].fix(0)
        m.b2.sb2.z_l[3].fix(0)

        vars_minus_x = (
            [m.b1.y, m.b2.z, m.x_f, m.b1.y_f, m.b2.z_f]
            + [m.b1.sb1.x_l[i] for i in indices]
            + [m.b1.sb1.y_l[i] for i in indices]
            + [m.b2.sb2.z_l[i] for i in indices]
        )

        m.con = pyo.Constraint(expr=sum(v for v in vars_minus_x) <= 1)
        m.b1.con = pyo.Constraint(expr=m.b1.y <= 1)
        m.b1.sb1.con = pyo.Constraint(expr=m.b1.sb1.y_l[0] <= 1)
        m.obj = pyo.Objective(expr=m.x)

        m.all_vars = ComponentSet([m.x] + vars_minus_x)
        m.unfixed_vars = ComponentSet([var for var in m.all_vars if not var.is_fixed()])

        return m

    def test_get_all_variables_unfixed(self):
        """Check that all unfixed variables are gathered."""
        m = self.get_var_model()
        var = au.get_model_variables(m)
        self.assertEqual(var, m.unfixed_vars)

    def test_get_all_variables(self):
        """Check that all fixed and unfixed variables are gathered."""
        m = self.get_var_model()
        var = au.get_model_variables(m, include_fixed=True)
        self.assertEqual(var, m.all_vars)

    def test_get_all_continuous(self):
        """Check that all continuous variables are gathered."""
        m = self.get_var_model()
        var = au.get_model_variables(
            m, include_continuous=True, include_binary=False, include_integer=False
        )
        continuous_vars = ComponentSet(
            var for var in m.unfixed_vars if var.is_continuous()
        )
        self.assertEqual(var, continuous_vars)

    def test_get_all_binary(self):
        """Check that all binary variables are gathered."""
        m = self.get_var_model()
        var = au.get_model_variables(
            m, include_continuous=False, include_binary=True, include_integer=False
        )
        binary_vars = ComponentSet(var for var in m.unfixed_vars if var.is_binary())
        self.assertEqual(var, binary_vars)

    def test_get_all_integer(self):
        """Check that all integer variables are gathered."""
        m = self.get_var_model()
        var = au.get_model_variables(
            m, include_continuous=False, include_binary=False, include_integer=True
        )
        continuous_vars = ComponentSet(
            var for var in m.unfixed_vars if var.is_integer()
        )
        self.assertEqual(var, continuous_vars)

    def test_get_specific_vars(self):
        """Check that all variables from a list are gathered."""
        m = self.get_var_model()
        components = [m.x, m.b1.sb1.y_l[0], m.b2.sb2.z_l]
        var = au.get_model_variables(m, components=components)
        specific_vars = ComponentSet(
            [m.x, m.b1.sb1.y_l[0], m.b2.sb2.z_l[0], m.b2.sb2.z_l[1], m.b2.sb2.z_l[2]]
        )
        self.assertEqual(var, specific_vars)

    def test_get_block_vars1(self):
        """
        Check that all variables from block are gathered (without
        descending into subblocks).
        """
        m = self.get_var_model()
        components = [m.b2.sb2.z_l, (m.b1, False)]
        var = au.get_model_variables(m, components=components)
        specific_vars = ComponentSet(
            [m.b1.y, m.b2.sb2.z_l[0], m.b2.sb2.z_l[1], m.b2.sb2.z_l[2]]
        )
        self.assertEqual(var, specific_vars)

    def test_get_block_vars2(self):
        """
        Check that all variables from block are gathered (without
        descending into subblocks).
        """
        m = self.get_var_model()
        components = [m.b1]
        var = au.get_model_variables(m, components=components)
        specific_vars = ComponentSet([m.b1.y, m.b1.sb1.y_l[0]])
        self.assertEqual(var, specific_vars)

    def test_get_constraint_vars(self):
        """Check that all variables constraints and objectives are gathered."""
        m = self.get_var_model()
        components = [m.con, m.obj]
        var = au.get_model_variables(m, components=components)
        self.assertEqual(var, m.unfixed_vars)


if __name__ == "__main__":
    unittest.main()



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________
#

import io
import pyomo.common.unittest as unittest
import pyomo.environ as pyo
from pyomo.opt.base.solvers import UnknownSolver
from pyomo.core.plugins.transform.scaling import ScaleModel, SuffixFinder
from pyomo.common.log import LoggingIntercept


class TestScaleModelTransformation(unittest.TestCase):
    def test_linear_scaling(self):
        model = pyo.ConcreteModel()
        model.x = pyo.Var([1, 2, 3], bounds=(-10, 10), initialize=5.0)
        model.z = pyo.Var(bounds=(10, 20))
        model.obj = pyo.Objective(expr=model.z + model.x[1])

        # test scaling of duals as well
        model.dual = pyo.Suffix(direction=pyo.Suffix.IMPORT)
        model.rc = pyo.Suffix(direction=pyo.Suffix.IMPORT)

        def con_rule(m, i):
            if i == 1:
                return m.x[1] + 2 * m.x[2] + 1 * m.x[3] == 4.0
            if i == 2:
                return m.x[1] + 2 * m.x[2] + 2 * m.x[3] == 5.0
            if i == 3:
                return m.x[1] + 3.0 * m.x[2] + 1 * m.x[3] == 5.0

        model.con = pyo.Constraint([1, 2, 3], rule=con_rule)
        model.zcon = pyo.Constraint(expr=model.z >= model.x[2])

        x_scale = 0.5
        obj_scale = 2.0
        z_scale = -10.0
        con_scale1 = 0.5
        con_scale2 = 2.0
        con_scale3 = -5.0
        zcon_scale = -3.0

        unscaled_model = model.clone()
        unscaled_model.scaling_factor = pyo.Suffix(direction=pyo.Suffix.EXPORT)
        unscaled_model.scaling_factor[unscaled_model.obj] = obj_scale
        unscaled_model.scaling_factor[unscaled_model.x] = x_scale
        unscaled_model.scaling_factor[unscaled_model.z] = z_scale
        unscaled_model.scaling_factor[unscaled_model.con[1]] = con_scale1
        unscaled_model.scaling_factor[unscaled_model.con[2]] = con_scale2
        unscaled_model.scaling_factor[unscaled_model.con[3]] = con_scale3
        unscaled_model.scaling_factor[unscaled_model.zcon] = zcon_scale

        scaled_model = pyo.TransformationFactory('core.scale_model').create_using(
            unscaled_model
        )

        # print('*** unscaled ***')
        # unscaled_model.pprint()
        # print('*** scaled ***')
        # scaled_model.pprint()

        glpk_solver = pyo.SolverFactory('glpk')
        if isinstance(glpk_solver, UnknownSolver) or (not glpk_solver.available()):
            raise unittest.SkipTest("glpk solver not available")

        glpk_solver.solve(unscaled_model)
        glpk_solver.solve(scaled_model)

        # check vars
        self.assertAlmostEqual(
            pyo.value(unscaled_model.x[1]),
            pyo.value(scaled_model.scaled_x[1]) / x_scale,
            4,
        )
        self.assertAlmostEqual(
            pyo.value(unscaled_model.x[2]),
            pyo.value(scaled_model.scaled_x[2]) / x_scale,
            4,
        )
        self.assertAlmostEqual(
            pyo.value(unscaled_model.x[3]),
            pyo.value(scaled_model.scaled_x[3]) / x_scale,
            4,
        )
        self.assertAlmostEqual(
            pyo.value(unscaled_model.z), pyo.value(scaled_model.scaled_z) / z_scale, 4
        )
        # check var lb
        self.assertAlmostEqual(
            pyo.value(unscaled_model.x[1].lb),
            pyo.value(scaled_model.scaled_x[1].lb) / x_scale,
            4,
        )
        self.assertAlmostEqual(
            pyo.value(unscaled_model.x[2].lb),
            pyo.value(scaled_model.scaled_x[2].lb) / x_scale,
            4,
        )
        self.assertAlmostEqual(
            pyo.value(unscaled_model.x[3].lb),
            pyo.value(scaled_model.scaled_x[3].lb) / x_scale,
            4,
        )
        # note: z_scale is negative, therefore, the inequality directions swap
        self.assertAlmostEqual(
            pyo.value(unscaled_model.z.lb),
            pyo.value(scaled_model.scaled_z.ub) / z_scale,
            4,
        )
        # check var ub
        self.assertAlmostEqual(
            pyo.value(unscaled_model.x[1].ub),
            pyo.value(scaled_model.scaled_x[1].ub) / x_scale,
            4,
        )
        self.assertAlmostEqual(
            pyo.value(unscaled_model.x[2].ub),
            pyo.value(scaled_model.scaled_x[2].ub) / x_scale,
            4,
        )
        self.assertAlmostEqual(
            pyo.value(unscaled_model.x[3].ub),
            pyo.value(scaled_model.scaled_x[3].ub) / x_scale,
            4,
        )
        # note: z_scale is negative, therefore, the inequality directions swap
        self.assertAlmostEqual(
            pyo.value(unscaled_model.z.ub),
            pyo.value(scaled_model.scaled_z.lb) / z_scale,
            4,
        )
        # check var multipliers (rc)
        self.assertAlmostEqual(
            pyo.value(unscaled_model.rc[unscaled_model.x[1]]),
            pyo.value(scaled_model.rc[scaled_model.scaled_x[1]]) * x_scale / obj_scale,
            4,
        )
        self.assertAlmostEqual(
            pyo.value(unscaled_model.rc[unscaled_model.x[2]]),
            pyo.value(scaled_model.rc[scaled_model.scaled_x[2]]) * x_scale / obj_scale,
            4,
        )
        self.assertAlmostEqual(
            pyo.value(unscaled_model.rc[unscaled_model.x[3]]),
            pyo.value(scaled_model.rc[scaled_model.scaled_x[3]]) * x_scale / obj_scale,
            4,
        )
        self.assertAlmostEqual(
            pyo.value(unscaled_model.rc[unscaled_model.z]),
            pyo.value(scaled_model.rc[scaled_model.scaled_z]) * z_scale / obj_scale,
            4,
        )
        # check constraint multipliers
        self.assertAlmostEqual(
            pyo.value(unscaled_model.dual[unscaled_model.con[1]]),
            pyo.value(scaled_model.dual[scaled_model.scaled_con[1]])
            * con_scale1
            / obj_scale,
            4,
        )
        self.assertAlmostEqual(
            pyo.value(unscaled_model.dual[unscaled_model.con[2]]),
            pyo.value(scaled_model.dual[scaled_model.scaled_con[2]])
            * con_scale2
            / obj_scale,
            4,
        )
        self.assertAlmostEqual(
            pyo.value(unscaled_model.dual[unscaled_model.con[3]]),
            pyo.value(scaled_model.dual[scaled_model.scaled_con[3]])
            * con_scale3
            / obj_scale,
            4,
        )

        # put the solution from the scaled back into the original
        pyo.TransformationFactory('core.scale_model').propagate_solution(
            scaled_model, model
        )

        # compare var values and rc with the unscaled soln
        for vm in model.component_objects(ctype=pyo.Var, descend_into=True):
            cuid = pyo.ComponentUID(vm)
            vum = cuid.find_component_on(unscaled_model)
            self.assertEqual((vm in model.rc), (vum in unscaled_model.rc))
            if vm in model.rc:
                self.assertAlmostEqual(
                    pyo.value(model.rc[vm]), pyo.value(unscaled_model.rc[vum]), 4
                )
            for k in vm:
                vmk = vm[k]
                vumk = vum[k]
                self.assertAlmostEqual(pyo.value(vmk), pyo.value(vumk), 4)
                self.assertEqual((vmk in model.rc), (vumk in unscaled_model.rc))
                if vmk in model.rc:
                    self.assertAlmostEqual(
                        pyo.value(model.rc[vmk]), pyo.value(unscaled_model.rc[vumk]), 4
                    )

        # compare constraint duals and value
        for model_con in model.component_objects(
            ctype=pyo.Constraint, descend_into=True
        ):
            cuid = pyo.ComponentUID(model_con)
            unscaled_model_con = cuid.find_component_on(unscaled_model)
            self.assertEqual(
                (model_con in model.rc), (unscaled_model_con in unscaled_model.rc)
            )
            if model_con in model.dual:
                self.assertAlmostEqual(
                    pyo.value(model.dual[model_con]),
                    pyo.value(unscaled_model.dual[unscaled_model_con]),
                    4,
                )
            for k in model_con:
                mk = model_con[k]
                umk = unscaled_model_con[k]
                self.assertEqual((mk in model.dual), (umk in unscaled_model.dual))
                if mk in model.dual:
                    self.assertAlmostEqual(
                        pyo.value(model.dual[mk]),
                        pyo.value(unscaled_model.dual[umk]),
                        4,
                    )

    def test_scaling_without_rename(self):
        m = pyo.ConcreteModel()
        m.scaling_factor = pyo.Suffix(direction=pyo.Suffix.EXPORT)
        m.v1 = pyo.Var(initialize=10)
        m.v2 = pyo.Var(initialize=20)
        m.v3 = pyo.Var(initialize=30)

        def c1_rule(m):
            return m.v1 == 1e6

        m.c1 = pyo.Constraint(rule=c1_rule)

        def c2_rule(m):
            return m.v2 == 1e-4

        m.c2 = pyo.Constraint(rule=c2_rule)

        m.scaling_factor[m.v1] = 1.0
        m.scaling_factor[m.v2] = 0.5
        m.scaling_factor[m.v3] = 0.25
        m.scaling_factor[m.c1] = 1e-5
        m.scaling_factor[m.c2] = 1e5

        values = {}
        values[id(m.v1)] = (m.v1.value, m.scaling_factor[m.v1])
        values[id(m.v2)] = (m.v2.value, m.scaling_factor[m.v2])
        values[id(m.v3)] = (m.v3.value, m.scaling_factor[m.v3])
        values[id(m.c1)] = (pyo.value(m.c1.body), m.scaling_factor[m.c1])
        values[id(m.c2)] = (pyo.value(m.c2.body), m.scaling_factor[m.c2])

        m.c2_ref = pyo.Reference(m.c2)
        m.v3_ref = pyo.Reference(m.v3)

        scale = pyo.TransformationFactory('core.scale_model')
        scale.apply_to(m, rename=False)

        self.assertTrue(hasattr(m, 'v1'))
        self.assertTrue(hasattr(m, 'v2'))
        self.assertTrue(hasattr(m, 'c1'))
        self.assertTrue(hasattr(m, 'c2'))

        orig_val, factor = values[id(m.v1)]
        self.assertAlmostEqual(m.v1.value, orig_val * factor)

        orig_val, factor = values[id(m.v2)]
        self.assertAlmostEqual(m.v2.value, orig_val * factor)

        orig_val, factor = values[id(m.c1)]
        self.assertAlmostEqual(pyo.value(m.c1.body), orig_val * factor)

        orig_val, factor = values[id(m.c2)]
        self.assertAlmostEqual(pyo.value(m.c2.body), orig_val * factor)

        orig_val, factor = values[id(m.v3)]
        self.assertAlmostEqual(m.v3_ref[None].value, orig_val * factor)
        # Note that because the model was not renamed,
        # v3_ref is still intact.

        lhs = m.c2.body
        monom_factor = lhs.arg(0)
        scale_factor = m.scaling_factor[m.c2] / m.scaling_factor[m.v2]
        self.assertAlmostEqual(monom_factor, scale_factor)

    def test_scaling_hierarchical(self):
        m = pyo.ConcreteModel()
        m.scaling_factor = pyo.Suffix(direction=pyo.Suffix.EXPORT)
        m.v1 = pyo.Var(initialize=10)
        m.v2 = pyo.Var(initialize=20)
        m.v3 = pyo.Var(initialize=30)

        def c1_rule(m):
            return m.v1 == 1e6

        m.c1 = pyo.Constraint(rule=c1_rule)

        def c2_rule(m):
            return m.v2 == 1e-4

        m.c2 = pyo.Constraint(rule=c2_rule)

        m.scaling_factor[m.v1] = 1.0
        m.scaling_factor[m.v2] = 0.5
        m.scaling_factor[m.v3] = 0.25
        m.scaling_factor[m.c1] = 1e-5
        m.scaling_factor[m.c2] = 1e5

        m.b = pyo.Block()
        m.b.scaling_factor = pyo.Suffix(direction=pyo.Suffix.EXPORT)
        m.b.v4 = pyo.Var(initialize=10)
        m.b.v5 = pyo.Var(initialize=20)
        m.b.v6 = pyo.Var(initialize=30)

        def c3_rule(m):
            return m.v4 == 1e6

        m.b.c3 = pyo.Constraint(rule=c3_rule)

        def c4_rule(m):
            return m.v5 == 1e-4

        m.b.c4 = pyo.Constraint(rule=c4_rule)

        m.b.scaling_factor[m.b.v4] = 1.0
        m.b.scaling_factor[m.b.v5] = 0.5
        m.b.scaling_factor[m.b.v6] = 0.25
        m.b.scaling_factor[m.b.c3] = 1e-5
        m.b.scaling_factor[m.b.c4] = 1e5

        values = {}
        values[id(m.v1)] = (m.v1.value, m.scaling_factor[m.v1])
        values[id(m.v2)] = (m.v2.value, m.scaling_factor[m.v2])
        values[id(m.v3)] = (m.v3.value, m.scaling_factor[m.v3])
        values[id(m.c1)] = (pyo.value(m.c1.body), m.scaling_factor[m.c1])
        values[id(m.c2)] = (pyo.value(m.c2.body), m.scaling_factor[m.c2])
        values[id(m.b.v4)] = (m.b.v4.value, m.b.scaling_factor[m.b.v4])
        values[id(m.b.v5)] = (m.b.v5.value, m.b.scaling_factor[m.b.v5])
        values[id(m.b.v6)] = (m.b.v6.value, m.b.scaling_factor[m.b.v6])
        values[id(m.b.c3)] = (pyo.value(m.b.c3.body), m.b.scaling_factor[m.b.c3])
        values[id(m.b.c4)] = (pyo.value(m.b.c4.body), m.b.scaling_factor[m.b.c4])

        m.c2_ref = pyo.Reference(m.c2)
        m.v3_ref = pyo.Reference(m.v3)

        m.b.c4_ref = pyo.Reference(m.b.c4)
        m.b.v6_ref = pyo.Reference(m.b.v6)

        scale = pyo.TransformationFactory('core.scale_model')
        scale.apply_to(m, rename=False)

        self.assertTrue(hasattr(m, 'v1'))
        self.assertTrue(hasattr(m, 'v2'))
        self.assertTrue(hasattr(m, 'c1'))
        self.assertTrue(hasattr(m, 'c2'))
        self.assertTrue(hasattr(m.b, 'v4'))
        self.assertTrue(hasattr(m.b, 'v5'))
        self.assertTrue(hasattr(m.b, 'c3'))
        self.assertTrue(hasattr(m.b, 'c4'))

        orig_val, factor = values[id(m.v1)]
        self.assertAlmostEqual(m.v1.value, orig_val * factor)

        orig_val, factor = values[id(m.v2)]
        self.assertAlmostEqual(m.v2.value, orig_val * factor)

        orig_val, factor = values[id(m.c1)]
        self.assertAlmostEqual(pyo.value(m.c1.body), orig_val * factor)

        orig_val, factor = values[id(m.c2)]
        self.assertAlmostEqual(pyo.value(m.c2.body), orig_val * factor)

        orig_val, factor = values[id(m.v3)]
        self.assertAlmostEqual(m.v3_ref[None].value, orig_val * factor)
        # Note that because the model was not renamed,
        # v3_ref is still intact.

        orig_val, factor = values[id(m.b.v4)]
        self.assertAlmostEqual(m.b.v4.value, orig_val * factor)

        orig_val, factor = values[id(m.b.v5)]
        self.assertAlmostEqual(m.b.v5.value, orig_val * factor)

        orig_val, factor = values[id(m.b.c3)]
        self.assertAlmostEqual(pyo.value(m.b.c3.body), orig_val * factor)

        orig_val, factor = values[id(m.b.c4)]
        self.assertAlmostEqual(pyo.value(m.b.c4.body), orig_val * factor)

        orig_val, factor = values[id(m.b.v6)]
        self.assertAlmostEqual(m.b.v6_ref[None].value, orig_val * factor)
        # Note that because the model was not renamed,
        # v6_ref is still intact.

        lhs = m.c2.body
        monom_factor = lhs.arg(0)
        scale_factor = m.scaling_factor[m.c2] / m.scaling_factor[m.v2]
        self.assertAlmostEqual(monom_factor, scale_factor)

        lhs = m.b.c4.body
        monom_factor = lhs.arg(0)
        scale_factor = m.b.scaling_factor[m.b.c4] / m.b.scaling_factor[m.b.v5]
        self.assertAlmostEqual(monom_factor, scale_factor)

    def test_scaling_no_solve(self):
        model = pyo.ConcreteModel()
        model.x = pyo.Var([1, 2, 3], bounds=(-10, 10), initialize=5.0)
        model.z = pyo.Var(bounds=(10, 20), initialize=15)

        def con_rule(m, i):
            if i == 1:
                return m.x[1] + 2 * m.x[2] + 1 * m.x[3] == 8.0
            if i == 2:
                return m.x[1] + 2 * m.x[2] + 2 * m.x[3] == 11.0
            if i == 3:
                return m.x[1] + 3.0 * m.x[2] + 1 * m.x[3] == 10.0

        model.con = pyo.Constraint([1, 2, 3], rule=con_rule)
        model.zcon = pyo.Constraint(expr=model.z >= model.x[2])

        model.x_ref = pyo.Reference(model.x)

        x_scale = 0.5
        obj_scale = 2.0
        z_scale = -10.0
        con_scale1 = 0.5
        con_scale2 = 2.0
        con_scale3 = -5.0
        zcon_scale = -3.0

        model.scaling_factor = pyo.Suffix(direction=pyo.Suffix.EXPORT)
        model.scaling_factor[model.x] = x_scale
        model.scaling_factor[model.z] = z_scale
        model.scaling_factor[model.con[1]] = con_scale1
        model.scaling_factor[model.con[2]] = con_scale2
        model.scaling_factor[model.con[3]] = con_scale3
        model.scaling_factor[model.zcon] = zcon_scale

        # Set scaling factors for References too, but these should be ignored by the transformation
        model.scaling_factor[model.x_ref] = x_scale * 2

        scaled_model = pyo.TransformationFactory('core.scale_model').create_using(model)

        # check vars
        self.assertAlmostEqual(
            pyo.value(model.x[1]), pyo.value(scaled_model.scaled_x[1]) / x_scale, 4
        )
        self.assertAlmostEqual(
            pyo.value(model.x[2]), pyo.value(scaled_model.scaled_x[2]) / x_scale, 4
        )
        self.assertAlmostEqual(
            pyo.value(model.x[3]), pyo.value(scaled_model.scaled_x[3]) / x_scale, 4
        )
        self.assertAlmostEqual(
            pyo.value(model.z), pyo.value(scaled_model.scaled_z) / z_scale, 4
        )
        # check var lb
        self.assertAlmostEqual(
            pyo.value(model.x[1].lb),
            pyo.value(scaled_model.scaled_x[1].lb) / x_scale,
            4,
        )
        self.assertAlmostEqual(
            pyo.value(model.x[2].lb),
            pyo.value(scaled_model.scaled_x[2].lb) / x_scale,
            4,
        )
        self.assertAlmostEqual(
            pyo.value(model.x[3].lb),
            pyo.value(scaled_model.scaled_x[3].lb) / x_scale,
            4,
        )
        # note: z_scale is negative, therefore, the inequality directions swap
        self.assertAlmostEqual(
            pyo.value(model.z.lb), pyo.value(scaled_model.scaled_z.ub) / z_scale, 4
        )
        # check var ub
        self.assertAlmostEqual(
            pyo.value(model.x[1].ub),
            pyo.value(scaled_model.scaled_x[1].ub) / x_scale,
            4,
        )
        self.assertAlmostEqual(
            pyo.value(model.x[2].ub),
            pyo.value(scaled_model.scaled_x[2].ub) / x_scale,
            4,
        )
        self.assertAlmostEqual(
            pyo.value(model.x[3].ub),
            pyo.value(scaled_model.scaled_x[3].ub) / x_scale,
            4,
        )
        # note: z_scale is negative, therefore, the inequality directions swap
        self.assertAlmostEqual(
            pyo.value(model.z.ub), pyo.value(scaled_model.scaled_z.lb) / z_scale, 4
        )

        # check references to vars
        self.assertAlmostEqual(
            pyo.value(model.x[1]), pyo.value(scaled_model.scaled_x_ref[1]) / x_scale, 4
        )
        self.assertAlmostEqual(
            pyo.value(model.x[2]), pyo.value(scaled_model.scaled_x_ref[2]) / x_scale, 4
        )
        self.assertAlmostEqual(
            pyo.value(model.x[3]), pyo.value(scaled_model.scaled_x_ref[3]) / x_scale, 4
        )
        # check var lb
        self.assertAlmostEqual(
            pyo.value(model.x[1].lb),
            pyo.value(scaled_model.scaled_x_ref[1].lb) / x_scale,
            4,
        )
        self.assertAlmostEqual(
            pyo.value(model.x[2].lb),
            pyo.value(scaled_model.scaled_x_ref[2].lb) / x_scale,
            4,
        )
        self.assertAlmostEqual(
            pyo.value(model.x[3].lb),
            pyo.value(scaled_model.scaled_x_ref[3].lb) / x_scale,
            4,
        )
        # note: z_scale is negative, therefore, the inequality directions swap
        self.assertAlmostEqual(
            pyo.value(model.z.lb), pyo.value(scaled_model.scaled_z.ub) / z_scale, 4
        )
        # check var ub
        self.assertAlmostEqual(
            pyo.value(model.x[1].ub),
            pyo.value(scaled_model.scaled_x_ref[1].ub) / x_scale,
            4,
        )
        self.assertAlmostEqual(
            pyo.value(model.x[2].ub),
            pyo.value(scaled_model.scaled_x_ref[2].ub) / x_scale,
            4,
        )
        self.assertAlmostEqual(
            pyo.value(model.x[3].ub),
            pyo.value(scaled_model.scaled_x_ref[3].ub) / x_scale,
            4,
        )

        # check constraints
        self.assertAlmostEqual(
            pyo.value(model.con[1]),
            pyo.value(scaled_model.scaled_con[1]) / con_scale1,
            4,
        )
        self.assertAlmostEqual(
            pyo.value(model.con[2]),
            pyo.value(scaled_model.scaled_con[2]) / con_scale2,
            4,
        )
        self.assertAlmostEqual(
            pyo.value(model.con[3]),
            pyo.value(scaled_model.scaled_con[3]) / con_scale3,
            4,
        )
        self.assertAlmostEqual(
            pyo.value(model.zcon), pyo.value(scaled_model.scaled_zcon) / zcon_scale, 4
        )

        # Set values on scaled model and check that they map back to original
        scaled_model.scaled_x[1].set_value(1 * x_scale)
        scaled_model.scaled_x[2].set_value(2 * x_scale)
        scaled_model.scaled_x[3].set_value(3 * x_scale)
        scaled_model.scaled_z.set_value(10 * z_scale)

        # put the solution from the scaled back into the original
        pyo.TransformationFactory('core.scale_model').propagate_solution(
            scaled_model, model
        )

        # Check var values
        self.assertAlmostEqual(pyo.value(model.x[1]), 1, 4)
        self.assertAlmostEqual(pyo.value(model.x[2]), 2, 4)
        self.assertAlmostEqual(pyo.value(model.x[3]), 3, 4)
        self.assertAlmostEqual(pyo.value(model.z), 10, 4)

        # Check reference values
        self.assertAlmostEqual(pyo.value(model.x_ref[1]), 1, 4)
        self.assertAlmostEqual(pyo.value(model.x_ref[2]), 2, 4)
        self.assertAlmostEqual(pyo.value(model.x_ref[3]), 3, 4)

        # check constraints
        self.assertAlmostEqual(pyo.value(model.con[1]), 8, 4)
        self.assertAlmostEqual(pyo.value(model.con[2]), 11, 4)
        self.assertAlmostEqual(pyo.value(model.con[3]), 10, 4)
        self.assertAlmostEqual(pyo.value(model.zcon), -8, 4)

    def test_get_float_scaling_factor_top_level(self):
        # Note: the transformation used to have a private method for
        # finding suffix values (which this method tested).  The
        # transformation now leverages the SuffixFinder.  To ensure that
        # the SuffixFinder behaves in the same way as the original local
        # method, we preserve these tests, but directly test the
        # SuffixFinder

        m = pyo.ConcreteModel()
        m.scaling_factor = pyo.Suffix(direction=pyo.Suffix.EXPORT)

        m.b1 = pyo.Block()
        m.b1.scaling_factor = pyo.Suffix(direction=pyo.Suffix.EXPORT)

        m.b1.b2 = pyo.Block()
        m.b1.b2.scaling_factor = pyo.Suffix(direction=pyo.Suffix.EXPORT)

        m.v1 = pyo.Var(initialize=10)
        m.b1.v2 = pyo.Var(initialize=20)
        m.b1.b2.v3 = pyo.Var(initialize=30)

        m.scaling_factor[m.v1] = 0.1
        m.scaling_factor[m.b1.v2] = 0.2

        _finder = SuffixFinder('scaling_factor', 1.0, m)

        # SF should be 0.1 from top level
        self.assertEqual(_finder.find(m.v1), 0.1)
        # SF should be 0.1 from top level, lower level ignored
        self.assertEqual(_finder.find(m.b1.v2), 0.2)
        # No SF, should return 1
        self.assertEqual(_finder.find(m.b1.b2.v3), 1.0)

    def test_get_float_scaling_factor_local_level(self):
        # Note: the transformation used to have a private method for
        # finding suffix values (which this method tested).  The
        # transformation now leverages the SuffixFinder.  To ensure that
        # the SuffixFinder behaves in the same way as the original local
        # method, we preserve these tests, but directly test the
        # SuffixFinder

        m = pyo.ConcreteModel()
        m.scaling_factor = pyo.Suffix(direction=pyo.Suffix.EXPORT)

        m.b1 = pyo.Block()
        m.b1.scaling_factor = pyo.Suffix(direction=pyo.Suffix.EXPORT)

        m.b1.b2 = pyo.Block()
        m.b1.b2.scaling_factor = pyo.Suffix(direction=pyo.Suffix.EXPORT)

        m.v1 = pyo.Var(initialize=10)
        m.b1.v2 = pyo.Var(initialize=20)
        m.b1.b2.v3 = pyo.Var(initialize=30)

        m.scaling_factor[m.v1] = 0.1
        m.b1.scaling_factor[m.b1.v2] = 0.2
        m.b1.b2.scaling_factor[m.b1.b2.v3] = 0.3

        # Add an intermediate scaling factor - this should take priority
        m.b1.scaling_factor[m.b1.b2.v3] = 0.4

        _finder = SuffixFinder('scaling_factor', 1.0, m)

        # Should get SF from local levels
        self.assertEqual(_finder.find(m.v1), 0.1)
        self.assertEqual(_finder.find(m.b1.v2), 0.2)
        self.assertEqual(_finder.find(m.b1.b2.v3), 0.4)

    def test_get_float_scaling_factor_intermediate_level(self):
        # Note: the transformation used to have a private method for
        # finding suffix values (which this method tested).  The
        # transformation now leverages the SuffixFinder.  To ensure that
        # the SuffixFinder behaves in the same way as the original local
        # method, we preserve these tests, but directly test the
        # SuffixFinder

        m = pyo.ConcreteModel()
        m.scaling_factor = pyo.Suffix(direction=pyo.Suffix.EXPORT)

        m.b1 = pyo.Block()
        m.b1.scaling_factor = pyo.Suffix(direction=pyo.Suffix.EXPORT)

        m.b1.b2 = pyo.Block()
        # No suffix at b2 level - this should not cause an issue

        m.b1.b2.b3 = pyo.Block()
        m.b1.b2.b3.scaling_factor = pyo.Suffix(direction=pyo.Suffix.EXPORT)

        m.v1 = pyo.Var(initialize=10)
        m.b1.b2.b3.v2 = pyo.Var(initialize=20)
        m.b1.b2.b3.v3 = pyo.Var(initialize=30)

        # Scale v1 at lowest level - this should not get picked up
        m.b1.b2.b3.scaling_factor[m.v1] = 0.1

        m.b1.scaling_factor[m.b1.b2.b3.v2] = 0.2
        m.b1.scaling_factor[m.b1.b2.b3.v3] = 0.3

        m.b1.b2.b3.scaling_factor[m.b1.b2.b3.v3] = 0.4

        _finder = SuffixFinder('scaling_factor', 1.0, m)

        # v1 should be unscaled as SF set below variable level
        self.assertEqual(_finder.find(m.v1), 1.0)
        # v2 should get SF from b1 level
        self.assertEqual(_finder.find(m.b1.b2.b3.v2), 0.2)
        # v2 should get SF from highest level, ignoring b3 level
        self.assertEqual(_finder.find(m.b1.b2.b3.v3), 0.3)

    def test_propagate_solution_uninitialized_variable(self):
        m = pyo.ConcreteModel()
        m.x = pyo.Var([1, 2], initialize=1.0)
        m.scaling_factor = pyo.Suffix(direction=pyo.Suffix.EXPORT)
        m.scaling_factor[m.x[1]] = 10.0
        m.scaling_factor[m.x[2]] = 10.0
        scaled_model = pyo.TransformationFactory("core.scale_model").create_using(m)
        scaled_model.scaled_x[1] = 20.0
        scaled_model.scaled_x[2] = None

        OUTPUT = io.StringIO()
        with LoggingIntercept(OUTPUT, "pyomo.core.plugins.transform.scaling"):
            pyo.TransformationFactory("core.scale_model").propagate_solution(
                scaled_model, m
            )
        msg = (
            "Variable with value None in the scaled model is replacing value of"
            " variable x[2] in the original model with None (was 1.0).\n"
        )
        self.assertEqual(OUTPUT.getvalue(), msg)
        self.assertAlmostEqual(m.x[1].value, 2.0, delta=1e-8)
        # Note that value of x[2] in original model *has* been overridden to None.
        # In this case, a warning has been raised.
        self.assertIs(m.x[2].value, None)


if __name__ == "__main__":
    unittest.main()
